"""Test filtering of the density field."""

import matplotlib as mpl
import numpy as np
import pytest
from matplotlib.colors import Normalize
from py21cmfast.c_21cmfast import ffi, lib
from scipy.stats import binned_statistic as binstat

from py21cmfast.wrapper.cfuncs import broadcast_input_struct

from . import produce_integration_test_data as prd
from .test_c_interpolation_tables import print_failure_stats

options_filter = [0, 1, 2, 3, 4]  # cell densities to draw samples from
R_PARAM_LIST = [
    1.5,
    5,
    10,
    20,
]  # default test HII_DIM = 50, we want max R < BOX_LEN*HII_DIM/3


# NOTE: These don't directly test against the expected FFT of these filters applied
#   to a central cell, but the continuous FT filters applied to a delta function.
#   this makes it a test of both the filter construction, as well as the aliasing.
# NOTE: These do not include the periodic boundary conditions, so face issues with R ~> HII_DIM/2.
def get_expected_output_centre(r_in, R_filter, R_param, filter_flag):
    # single pixel boxes have a specific shape based on the filter
    R_ratio = r_in / R_filter
    if filter_flag == 0:
        # output is uniform sphere around the centre point
        exp_vol = 4 / 3 * np.pi * R_filter**3
        return (R_ratio < 1) / exp_vol
    elif filter_flag == 1:
        # output is the tophat FFT
        R_ratio *= 1 / 0.413566994  # == this is the 2*pi*k factor for equating volumes
        result = (np.sin(R_ratio) - R_ratio * np.cos(R_ratio)) / (
            2 * np.pi**2 * r_in**3
        )
        result[r_in == 0] = (
            1 / 6 / np.pi**2 * (0.413566994 * R_filter) ** 3
        )  # r->0 limit
        return result
    elif filter_flag == 2:
        # output is Gaussian
        const = (0.643 * R_filter) ** 2
        exp_vol = (2 * np.pi * const) ** (3.0 / 2.0)
        return np.exp(-((r_in) ** 2 / const / 2)) / exp_vol
    elif filter_flag == 3:
        # output is sphere with exponential damping
        exp_vol = 4 / 3 * np.pi * R_filter**3
        return (R_ratio < 1) * np.exp(-r_in / R_param) / exp_vol
    elif filter_flag == 4:
        # output is spherical shell
        exp_vol = 4 / 3 * np.pi * (R_filter**3 - R_param**3)
        return (R_ratio < 1) * (R_param <= r_in) / exp_vol


# return binned quantities
def get_binned_stats(x_arr, y_arr, bins, stats):
    x_in = x_arr.flatten()
    y_in = y_arr.flatten()
    result = {}

    statistic_dict = {
        "pc1u": lambda x: np.percentile(x, 84),
        "pc1l": lambda x: np.percentile(x, 16),
        "pc2u": lambda x: np.percentile(x, 97.5),
        "pc2l": lambda x: np.percentile(x, 2.5),
        # used to mark percentiles in errorbar plots, since these cannot be negative
        "err1u": lambda x: np.maximum(np.percentile(x, 84) - np.mean(x), 0),
        "err2u": lambda x: np.maximum(np.percentile(x, 97.5) - np.mean(x), 0),
        "err1l": lambda x: np.maximum(np.mean(x) - np.percentile(x, 16), 0),
        "err2l": lambda x: np.maximum(np.mean(x) - np.percentile(x, 2.5), 0),
        "errmin": lambda x: np.mean(x) - np.amin(x),
        "errmax": lambda x: np.amax(x) - np.mean(x),
    }

    for stat in stats:
        spstatkey = statistic_dict.get(stat, stat)
        result[stat], _, _ = binstat(x_in, y_in, bins=bins, statistic=spstatkey)

    return result


@pytest.mark.parametrize("R", R_PARAM_LIST)
@pytest.mark.parametrize("filter_flag", options_filter)
def test_filters(filter_flag, R, plt):
    opts = prd.get_all_options_struct(redshift=10.0)

    inputs = opts["inputs"]
    up = inputs.simulation_options

    # testing a single pixel source
    input_box_centre = np.zeros((up.HII_DIM,) * 3, dtype="f4")
    input_box_centre[up.HII_DIM // 2, up.HII_DIM // 2, up.HII_DIM // 2] = 1.0
    output_box_centre = np.zeros((up.HII_DIM,) * 3, dtype="f8")
    # use MFP=20 for the exp filter, use a 4 cell shell for the annular filter
    if filter_flag == 3:
        R_param = 20
    elif filter_flag == 4:
        R_param = max(R - 4 * (up.BOX_LEN / up.HII_DIM), 0)
    else:
        R_param = 0

    broadcast_input_struct(inputs)
    lib.test_filter(
        ffi.cast("float *", input_box_centre.ctypes.data),
        R,
        R_param,
        filter_flag,
        ffi.cast("double *", output_box_centre.ctypes.data),
    )

    # expected outputs given in cell units
    R_cells = R / up.BOX_LEN * up.HII_DIM
    Rp_cells = R_param / up.BOX_LEN * up.HII_DIM
    r_from_centre = np.linalg.norm(
        np.mgrid[0 : up.HII_DIM, 0 : up.HII_DIM, 0 : up.HII_DIM]
        - np.array([up.HII_DIM // 2, up.HII_DIM // 2, up.HII_DIM // 2])[
            :, None, None, None
        ],
        axis=0,
    )
    # prevent divide by zero in the central cell
    r_from_centre[up.HII_DIM // 2, up.HII_DIM // 2, up.HII_DIM // 2] = 1e-6

    exp_output_centre = get_expected_output_centre(
        r_from_centre, R_cells, Rp_cells, filter_flag
    )

    # these are very wide tolerances, just to make sure there aren't
    # cells 100x the expected values
    abs_tol_pixel = exp_output_centre.max() * 0.8
    rel_tol_pixel = 0
    # we take bins of 2 pixels to smooth over sharp edged filters
    r_bins = np.arange(0, int(up.HII_DIM / 2 * np.sqrt(3)), 2)
    r_cen = (r_bins[1:] + r_bins[:-1]) / 2

    binned_truth_centre = get_binned_stats(
        r_from_centre,
        exp_output_centre,
        r_bins,
        stats=["mean", "errmin", "errmax", "err1l", "err1u"],
    )
    binned_truth_centre = binned_truth_centre["mean"]

    stats_o = get_binned_stats(
        r_from_centre,
        output_box_centre,
        r_bins,
        stats=["mean", "errmin", "errmax", "err1l", "err1u"],
    )

    if plt == mpl.pyplot:
        filter_plot(
            inputs=[input_box_centre],
            outputs=[output_box_centre],
            binned_truths=[binned_truth_centre],
            binned_stats=[stats_o],
            truths=[exp_output_centre],
            r_bins=r_bins,
            r_grid=r_from_centre,
            slice_index=up.HII_DIM // 2,
            slice_axis=0,
            abs_tol=abs_tol_pixel,
            rel_tol=rel_tol_pixel,
            plt=plt,
        )

    # All filters should be normalised aside from the exp filter
    if filter_flag == 3:
        # ratio of exponential and sphere volume integrals
        R_q = R_param / R
        norm_factor = 6 * R_q**3 - np.exp(-1 / R_q) * (
            6 * R_q**3 + 6 * R_q**2 + 3 * R_q
        )
    else:
        norm_factor = 1
    # firstly, make sure the filters are normalised
    np.testing.assert_allclose(
        input_box_centre.sum() * norm_factor, output_box_centre.sum(), atol=1e-4
    )

    # secondly, make sure binned results are reasonable
    rel_tol_bins = 1e-1
    abs_tol_bins = exp_output_centre.max() * 1e-1
    print_failure_stats(
        stats_o["mean"],
        binned_truth_centre,
        [r_cen],
        abs_tol=abs_tol_bins,
        rel_tol=rel_tol_bins,
        name="bins",
    )
    np.testing.assert_allclose(
        binned_truth_centre, stats_o["mean"], atol=abs_tol_bins, rtol=rel_tol_bins
    )

    # lastly, make sure no pixels are way out of line.
    # this has a wide tolerance due to significant aliasing
    print_failure_stats(
        output_box_centre,
        exp_output_centre,
        [r_from_centre],
        abs_tol=abs_tol_pixel,
        rel_tol=rel_tol_pixel,
        name="pixel",
    )
    np.testing.assert_allclose(
        output_box_centre, exp_output_centre, rtol=rel_tol_pixel, atol=abs_tol_pixel
    )


# since the filters are symmetric I'm doing an R vs value plot instead of imshowing slices
def filter_plot(
    inputs,
    outputs,
    binned_truths,
    binned_stats,
    truths,
    r_bins,
    r_grid,
    slice_index,
    slice_axis,
    abs_tol,
    rel_tol,
    plt,
):
    if not (len(inputs) == len(binned_truths) == len(outputs)):
        raise ValueError(
            f"inputs {len(inputs)}, outputs {len(outputs)} and truths {len(binned_truths)}"
            "must have the same length."
        )

    n_plots = len(inputs)
    fig, axs = plt.subplots(
        n_plots,
        4,
        figsize=(16.0, 12.0 * n_plots / 4.0),
        layout="constrained",
        squeeze=False,
    )
    fig.get_layout_engine().set(w_pad=0 / 72, h_pad=0 / 72, hspace=0.0, wspace=0.0)

    r_cen = (r_bins[1:] + r_bins[:-1]) / 2

    axs[0, 0].set_title("Output")
    axs[0, 1].set_title("Expected")
    axs[0, 2].set_title("Radii")
    axs[0, 3].set_title("Pixels")
    for idx, (_i, o, bo, t, tt) in enumerate(
        zip(inputs, outputs, binned_stats, binned_truths, truths, strict=True)
    ):
        axs[idx, 0].pcolormesh(
            o.take(indices=slice_index, axis=slice_axis),
            norm=Normalize(vmin=0, vmax=o.max()),
        )
        axs[idx, 1].pcolormesh(
            tt.take(indices=slice_index, axis=slice_axis),
            norm=Normalize(vmin=0, vmax=o.max()),
        )

        lns = []
        lns.append(
            axs[idx, 2].errorbar(
                r_cen,
                bo["mean"],
                markerfacecolor="b",
                elinewidth=1,
                capsize=3,
                markersize=5,
                marker="o",
                color="b",
                yerr=[bo["errmin"], bo["errmax"]],
                label="filtered grid",
                zorder=2,
            )
        )

        lns.append(
            axs[idx, 2].plot(r_cen, t, "m:", linewidth=2, label="Expected", zorder=3)[0]
        )
        axs[idx, 2].grid()
        axs[idx, 2].set_xlabel("dist from centre")
        axs[idx, 2].set_ylabel("cell value")
        axs[idx, 2].legend()

        err_base = np.linspace(tt.min(), tt.max(), num=100)
        err_max = err_base + (abs_tol + np.fabs(err_base) * rel_tol)
        err_min = err_base - (abs_tol + np.fabs(err_base) * rel_tol)
        axs[idx, 3].plot(err_base, err_max, "k:")
        axs[idx, 3].plot(err_base, err_min, "k:")
        axs[idx, 3].plot(err_base, err_base, "k--")

        axs[idx, 3].scatter(tt, o, s=1, alpha=0.5, rasterized=True)
        axs[idx, 3].grid()
        axs[idx, 3].set_xlabel("expected cell value")
        axs[idx, 3].set_ylabel("filtered cell value")



"""Tests of the input_serialization module."""

from typing import Literal

import pytest

from py21cmfast import InputParameters
from py21cmfast import input_serialization as srlz


class TestConvertInputsToDict:
    """Test the convert_inputs_to_dict function."""

    def test_default_minimal(self):
        """Test that ps_norm and USE_SIGMA_8 are in default inputs."""
        inputs = InputParameters(random_seed=0)
        out = srlz.convert_inputs_to_dict(inputs, mode="minimal")
        assert "CosmoTables" in out
        assert "ps_norm" in out["CosmoTables"]
        assert "USE_SIGMA_8" in out["CosmoTables"]

    @pytest.mark.parametrize("mode", ["full", "minimal"])
    def test_default_with_nonstructs(self, mode):
        """Test also reading extra info like random_seed."""
        inputs = InputParameters(random_seed=42)
        out = srlz.convert_inputs_to_dict(inputs, mode=mode, only_structs=False)
        assert out["random_seed"] == 42

    def test_default_full_snake(self):
        """Test reading in snake_case mode."""
        inputs = InputParameters(random_seed=42)
        out = srlz.convert_inputs_to_dict(inputs, mode="full", camel=False)
        assert "cosmo_params" in out

    def test_not_use_aliases(self):
        """Test not using aliases."""
        inputs = InputParameters(random_seed=42)
        out = srlz.convert_inputs_to_dict(inputs, mode="full", use_aliases=False)
        assert "_DIM" in out["SimulationOptions"]
        assert "DIM" not in out["SimulationOptions"]


class TestPrepareInputsForSerialization:
    """Tests of the prepare_inputs_for_serialization function."""

    def test_default_minimal(self):
        """Test that constructing a minimal serialization results in a dictionary that still contains ps_norm and USE_SIGMA_8."""
        inputs = InputParameters(random_seed=1)
        out = srlz.prepare_inputs_for_serialization(inputs, mode="minimal")
        assert "CosmoTables" in out
        assert "ps_norm" in out["CosmoTables"]
        assert "USE_SIGMA_8" in out["CosmoTables"]

    @pytest.mark.parametrize(
        "inputs",
        [
            InputParameters(random_seed=0),
            InputParameters.from_template("Park19", random_seed=0),
            InputParameters.from_template(
                "default", HII_DIM=50, DIM=100, BOX_LEN=50, random_seed=0
            ),
        ],
        ids=[
            "default",
            "park19",
            "explicit-dim-boxlen",
        ],
    )
    @pytest.mark.parametrize("mode", ["full", "minimal"])
    @pytest.mark.parametrize("camel", [True, False])
    def test_roundtrip(
        self, inputs: InputParameters, mode: Literal["full", "minimal"], camel: bool
    ):
        """Test that writing then reading gives back the same thing."""
        dct = srlz.prepare_inputs_for_serialization(
            inputs, mode=mode, only_structs=True, camel=camel
        )
        # Here we would generally write the output to some format like TOML, then
        # read that back in as a dict, then we need to deserialize from dict:
        dct = srlz.deserialize_inputs(dct)
        new = InputParameters(random_seed=inputs.random_seed, **dct)
        assert new == inputs


class TestDeserializeInputs:
    """Tests of the deserialize_inputs function."""

    def test_extra_toplevel_param_warns(self):
        """Test that extra top-level parameters yields a warning."""
        inputs = srlz.prepare_inputs_for_serialization(
            InputParameters(random_seed=0), mode="full"
        )
        inputs["extra_key"] = {}

        with pytest.warns(UserWarning, match="The following keys were not recognized"):
            srlz.deserialize_inputs(inputs)

    def test_extra_loose_param_warns(self):
        """Test that extra loose params that don't exist give warnings."""
        inputs = srlz.prepare_inputs_for_serialization(
            InputParameters(random_seed=0), mode="full"
        )

        with pytest.warns(UserWarning, match="Excess arguments"):
            srlz.deserialize_inputs(inputs, MY_EXTRA_PARAM=True, safe=False)



"""Modified YAML that can load/dump `astropy` quantities.

A modification of the basic YAML and astropy.io.misc.yaml to be able to load/dump
objects with astropy quantities in them.
"""

import yaml
from astropy.io.misc import yaml as ayaml


class _NewDumper(yaml.Dumper, ayaml.AstropyDumper):
    pass


class _NewLoader(yaml.Loader, ayaml.AstropyLoader):
    pass


for k, v in yaml.Dumper.yaml_representers.items():
    _NewDumper.add_representer(k, v)

for k, v in yaml.Dumper.yaml_multi_representers.items():
    _NewDumper.add_multi_representer(k, v)

for k, v in ayaml.AstropyDumper.yaml_representers.items():
    _NewDumper.add_representer(k, v)

for k, v in ayaml.AstropyDumper.yaml_multi_representers.items():
    _NewDumper.add_multi_representer(k, v)

for k, v in yaml.Loader.yaml_constructors.items():
    _NewLoader.add_constructor(k, v)

for k, v in ayaml.AstropyLoader.yaml_constructors.items():
    _NewLoader.add_constructor(k, v)

for k, v in yaml.Loader.yaml_multi_constructors.items():
    _NewLoader.add_multi_constructor(k, v)

for k, v in ayaml.AstropyLoader.yaml_multi_constructors.items():
    _NewLoader.add_multi_constructor(k, v)


def load(stream):
    """Load an object from a YAML stream."""
    return yaml.load(stream, Loader=_NewLoader)


def dump(data, stream=None, **kwargs):
    """Dump an object into a YAML stream."""
    kwargs["Dumper"] = _NewDumper
    return yaml.dump(data, stream=stream, **kwargs)



"""A module for classes that create lightcone slices from Coeval objects."""

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence
from functools import cached_property, partial

import attrs
import numpy as np
from astropy import units
from astropy.cosmology import FLRW, z_at_value
from astropy.units import Mpc, Quantity, pixel, pixel_scale
from classy import Class
from cosmotile import (
    make_lightcone_slice_interpolator,
    make_lightcone_slice_vector_field,
)
from scipy.spatial.transform import Rotation

from .rsds import estimate_rsd_displacements
from .wrapper.classy_interface import run_classy
from .wrapper.inputs import (
    InputParameters,
    Planck18,  # Not *quite* the same as astropy's Planck18
    SimulationOptions,
)

_LIGHTCONERS = {}
_LENGTH = "length"


@attrs.define(kw_only=True, slots=False)
class Lightconer(ABC):
    """A class that creates lightcone slices from Coeval objects.

    Parameters
    ----------
    lc_distances
        The comoving distances to the lightcone slices, in Mpc. Either this or
        the ``lc_redshifts`` must be provided.
    lc_redshifts
        The redshifts of the lightcone slices. Either this or the ``lc_distances``
        must be provided.
    cosmo
        The cosmology to use. Defaults to Planck18.
    quantities
        An iteratable of quantities to include in the lightcone slices. These should
        be attributes of the :class:~`outputs.Coeval` class that are arrays of
        shape ``HII_DIM^3``. A *special* value here is `velocity_los`, which  Defaults to ``("brightness_temp",)``.
    """

    cosmo: FLRW = attrs.field(
        default=Planck18, validator=attrs.validators.instance_of(FLRW)
    )

    _lc_redshifts: np.ndarray = attrs.field(default=None, eq=False)
    lc_distances: Quantity[_LENGTH] = attrs.field(
        eq=attrs.cmp_using(eq=partial(np.allclose, rtol=1e-5, atol=0))
    )

    quantities: tuple[str] = attrs.field(
        default=("brightness_temp",),
        converter=tuple,
        validator=attrs.validators.deep_iterable(attrs.validators.instance_of(str)),
    )
    interp_kinds: dict[str, str] = attrs.field()

    @lc_distances.default
    def _lcd_default(self):
        if self._lc_redshifts is None:
            raise ValueError("Either lc_distances or lc_redshifts must be provided")
        return self.cosmo.comoving_distance(self._lc_redshifts)

    @lc_distances.validator
    def _lcd_vld(self, attribute, value):
        if np.any(value < 0):
            raise ValueError("lc_distances must be non-negative")

    @_lc_redshifts.validator
    def _lcz_vld(self, attribute, value):
        if value is None:
            return

        if np.any(value < 0):
            raise ValueError("lc_redshifts must be non-negative")

    @cached_property
    def lc_redshifts(self) -> np.ndarray:
        """The redshifts of the lightcone slices."""
        if self._lc_redshifts is not None:
            return self._lc_redshifts

        d = self.lc_distances
        zmin = z_at_value(self.cosmo.comoving_distance, d.min()).value
        zmax = z_at_value(self.cosmo.comoving_distance, d.max()).value

        zgrid = np.logspace(np.log10(zmin), np.log10(zmax), 100)
        dgrid = self.cosmo.comoving_distance(zgrid)
        return np.interp(d.value, dgrid.value, zgrid)

    def get_lc_distances_in_pixels(self, resolution: Quantity[_LENGTH]):
        """Get the lightcone distances in pixels, given a resolution."""
        return self.lc_distances.to(pixel, pixel_scale(resolution / pixel))

    @interp_kinds.default
    def _interp_kinds_def(self):
        return {"z_reion": "mean_max"}

    def get_shape(self, simulation_options: SimulationOptions) -> tuple[int, int, int]:
        """Get the shape of the lightcone slices."""
        raise NotImplementedError

    @classmethod
    def between_redshifts(
        cls,
        min_redshift: float,
        max_redshift: float,
        resolution: Quantity[_LENGTH],
        cosmo=Planck18,
        **kw,
    ):
        """Construct a Lightconer with regular comoving dist. slices between two z's."""
        d_at_redshift = cosmo.comoving_distance(min_redshift).to_value(Mpc)
        dmax = cosmo.comoving_distance(max_redshift).to_value(Mpc)
        res = resolution.to_value(Mpc)

        lc_distances = np.arange(d_at_redshift, dmax + res, res)

        return cls(lc_distances=lc_distances * Mpc, cosmo=cosmo, **kw)

    @classmethod
    def with_equal_cdist_slices(
        cls,
        min_redshift: float,
        max_redshift: float,
        resolution: Quantity[_LENGTH],
        cosmo=Planck18,
        **kw,
    ):
        """
        Create a lightconer with equally spaced slices in comoving distance.

        This method is deprecated and will be removed in future versions.
        Instead, use `between_redshifts` to create a lightconer with equally spaced
        slices in comoving distance.
        """
        warnings.warn(
            "with_equal_cdist_slices is deprecated and will be removed in future versions. "
            "Call between_redshifts instead to silence this warning.",
            stacklevel=2,
        )
        return cls.between_redshifts(
            min_redshift=min_redshift,
            max_redshift=max_redshift,
            resolution=resolution,
            cosmo=cosmo,
            **kw,
        )

    def make_lightcone_slices(
        self,
        c1: Coeval,  # noqa: F821
        c2: Coeval,  # noqa: F821
    ) -> tuple[dict[str, np.ndarray], np.ndarray]:
        """
        Make lightcone slices out of two coeval objects.

        Parameters
        ----------
        c1, c2 : Coeval
            The coeval boxes to interpolate.

        Returns
        -------
        quantity
            The field names of the quantities required by the lightcone.
        lcidx
            The indices of the lightcone to which these slices belong.
        scalar_field_slices
            The scalar fields evaluated on the "lightcone" slices that exist within
            the redshift range spanned by ``c1`` and ``c2``.
        """
        if c1.simulation_options != c2.simulation_options:
            raise ValueError("c1 and c2 must have the same user parameters")
        if c1.cosmo_params != c2.cosmo_params:
            raise ValueError("c1 and c2 must have the same cosmological parameters")

        cosmo = c1.cosmo_params.cosmo
        pixeleq = pixel_scale(c1.simulation_options.cell_size / pixel)

        dc1 = cosmo.comoving_distance(c1.redshift).to(pixel, equivalencies=pixeleq)
        dc2 = cosmo.comoving_distance(c2.redshift).to(pixel, equivalencies=pixeleq)

        dcmin = min(dc1, dc2)
        dcmax = max(dc1, dc2)

        pixlcdist = self.get_lc_distances_in_pixels(c1.simulation_options.cell_size)

        # At the lower redshift, we include some tolerance. This is because the very
        # last slice (lowest redshift) may correspond *exactly* to the lowest coeval
        # box, and due to rounding error in the `z_at_value` call, they might be
        # slightly off.
        lcidx = np.nonzero((pixlcdist >= dcmin * (1 - 1e-6)) & (pixlcdist < dcmax))[0]

        # Return early if no lightcone indices are between the coeval distances.
        if len(lcidx) == 0:
            yield None, lcidx, None

        lc_distances = pixlcdist[lcidx]

        lc_quantities = self.quantities

        for idx, lcd in zip(lcidx, lc_distances, strict=True):
            for q in lc_quantities:
                if q == "los_velocity":
                    continue
                box1 = self.coeval_subselect(
                    lcd, getattr(c1, q), c1.simulation_options.cell_size
                )
                box2 = self.coeval_subselect(
                    lcd, getattr(c2, q), c2.simulation_options.cell_size
                )
                box = self.redshift_interpolation(
                    lcd, box1, box2, dc1, dc2, kind=self.interp_kinds.get(q, "mean")
                )

                yield q, idx, self.construct_lightcone(lcd, box)

                if "los_velocity" in self.quantities and q == self.quantities[0]:
                    # While doing the first quantity, also add in the los velocity, if desired.
                    # Doing it now means we can keep whatever cached interpolation setup
                    # is used to do construct_lightcone().
                    if isinstance(self, AngularLightconer):
                        boxes1 = [
                            self.coeval_subselect(
                                lcd,
                                getattr(c1, f"velocity_{q}"),
                                c1.simulation_options.cell_size,
                            )
                            for q in "xyz"
                        ]
                        boxes2 = [
                            self.coeval_subselect(
                                lcd,
                                getattr(c2, f"velocity_{q}"),
                                c2.simulation_options.cell_size,
                            )
                            for q in "xyz"
                        ]

                        interpolated_boxes = [
                            self.redshift_interpolation(
                                lcd,
                                box1,
                                box2,
                                dc1,
                                dc2,
                                kind=self.interp_kinds.get("velocity", "mean"),
                            )
                            for (box1, box2) in zip(boxes1, boxes2, strict=True)
                        ]
                    else:
                        # TODO: get the correct component according to self.line_of_sight_axis
                        box1 = self.coeval_subselect(
                            lcd, c1.velocity_z, c1.simulation_options.cell_size
                        )
                        box2 = self.coeval_subselect(
                            lcd, c2.velocity_z, c2.simulation_options.cell_size
                        )
                        interpolated_boxes = self.redshift_interpolation(
                            lcd,
                            box1,
                            box2,
                            dc1,
                            dc2,
                            kind=self.interp_kinds.get("velocity_z", "mean"),
                        )
                    yield (
                        "los_velocity",
                        idx,
                        self.construct_los_velocity_lightcone(
                            lcd,
                            interpolated_boxes,
                        ),
                    )

    def coeval_subselect(
        self, lcd: float, coeval: np.ndarray, coeval_res: Quantity[_LENGTH]
    ) -> np.ndarray:
        """Sub-Select the coeval box required for interpolation at one slice."""
        return coeval

    def redshift_interpolation(
        self,
        dc: float,
        coeval_a: np.ndarray,
        coeval_b: np.ndarray,
        dc_a: float,
        dc_b: float,
        kind: str = "mean",
    ) -> np.ndarray:
        """Perform redshift interpolation to a new box given two bracketing coevals."""
        if coeval_a.shape != coeval_b.shape:
            raise ValueError("coeval_a and coeval_b must have the same shape")

        out = (np.abs(dc_b - dc) * coeval_a + np.abs(dc_a - dc) * coeval_b) / np.abs(
            dc_a - dc_b
        )

        if kind == "mean_max":
            flag = coeval_a * coeval_b < 0
            out[flag] = np.maximum(coeval_a, coeval_b)[flag]
        elif kind != "mean":
            raise ValueError("kind must be 'mean' or 'mean_max'")

        return out

    @abstractmethod
    def construct_lightcone(
        self,
        lc_distances: np.ndarray,
        boxes: Sequence[np.ndarray],
    ) -> np.ndarray:
        """Abstract method for constructing the lightcone slices."""

    @abstractmethod
    def construct_los_velocity_lightcone(
        self,
        lc_distances: np.ndarray,
        velocities: tuple[np.ndarray, np.ndarray, np.ndarray],
    ) -> np.ndarray:
        """Abstract method for constructing the LoS velocity lightcone slices."""

    def validate_options(
        self,
        inputs: InputParameters,
        include_dvdr_in_tau21: bool,
        apply_rsds: bool,
        classy_output: Class | None = None,
    ) -> Lightconer:
        """Validate 21cmFAST options."""
        if len(inputs.node_redshifts) == 0:
            raise ValueError(
                "You are attempting to run a lightcone with no node_redshifts."
                "Please set node_redshifts on the `inputs` parameter."
            )
        lcd = self.lc_distances
        d_node_min = self.cosmo.comoving_distance(min(inputs.node_redshifts))
        d_node_max = self.cosmo.comoving_distance(max(inputs.node_redshifts))
        if not ((d_node_min <= lcd.min()) and (lcd.max() <= d_node_max)):
            lcz = self.lc_redshifts
            raise ValueError(
                "The lightcone redshifts are not compatible with the given redshift."
                f"The range of computed redshifts is {min(inputs.node_redshifts)} to {max(inputs.node_redshifts)}, "
                f"while the lightcone redshift range is {lcz.min()} to {lcz.max()}. "
                "Extend the limits of node redshifts to avoid this error."
            )

        if (
            include_dvdr_in_tau21
            and inputs.astro_options.USE_TS_FLUCT
            and "tau_21" not in self.quantities
        ):
            lightconer = attrs.evolve(self, quantities=(*self.quantities, "tau_21"))
        else:
            lightconer = self

        if apply_rsds or include_dvdr_in_tau21:
            lightconer = attrs.evolve(
                lightconer, quantities=(*lightconer.quantities, "los_velocity")
            )

        if not apply_rsds:
            return lightconer

        if classy_output is None:
            classy_output = run_classy(
                h=inputs.cosmo_params.hlittle,
                Omega_cdm=inputs.cosmo_params.OMm - inputs.cosmo_params.OMb,
                Omega_b=inputs.cosmo_params.OMb,
                n_s=inputs.cosmo_params.POWER_INDEX,
                sigma8=inputs.cosmo_params.SIGMA_8,
                output="vTk",
            )
        lcd_limits_rsd = self.find_required_lightcone_limits(
            classy_output=classy_output, inputs=inputs
        )
        lcd_rsd = (
            np.arange(
                lcd_limits_rsd[0].value,
                lcd_limits_rsd[1].value + inputs.simulation_options.cell_size.value,
                inputs.simulation_options.cell_size.value,
            )
            * self.lc_distances.unit
        )
        # Make a new lightconer which is identical to self, but extends to further boundaries!
        return attrs.evolve(lightconer, lc_redshifts=None, lc_distances=lcd_rsd)

    def find_required_lightcone_limits(
        self, classy_output: Class, inputs: InputParameters
    ) -> list[units.Quantity]:
        """Obtain the redshift limits required for the lightcone to include RSDs.

        This is a *crude* estimation of the maximum/minimum lightcone limits that are
        required in order to simulate all the "mass" that enters the requested
        ligthcone (due to RSD shift). We use the rms of the velocity field from linear
        perturbation theory in order to determine the required lightcone limits.
        If no limit is found, it means that the limits of node_redshifts are not
        sufficient and an error is raised.

        Parameters
        ----------
        classy_output : classy.Class
            An object containing all the information from the CLASS calculation.
        inputs: InputParameters
            The input parameters corresponding to the box.

        Returns
        -------
        lcd_limits_rsd : list
            List that contains the limits of the required lightcone distances.
        """
        # The velocity rms that CLASS returns corresponds to the magnitude of the velocity vector.
        # In case of rectilinear lightcones, since we shift the cells only along one axis (e.g. z-axis),
        # we need the rms of the correponding component of the velocity vector, which is smaller by sqrt(3)
        # due to isotropy
        factor = 1.0 if isinstance(self, AngularLightconer) else 1.0 / np.sqrt(3.0)
        z_node_limits = [min(inputs.node_redshifts), max(inputs.node_redshifts)]
        lcd_limits = [self.lc_distances.min(), self.lc_distances.max()]
        signs = [-1, 1]
        lcd_limits_rsd = []
        for sign, z_node_limit, lcd_limit in zip(
            signs, z_node_limits, lcd_limits, strict=False
        ):
            distances_out = (
                np.arange(
                    lcd_limit.value,
                    self.cosmo.comoving_distance(z_node_limit).value,
                    inputs.simulation_options.cell_size.value * sign,
                )
                * lcd_limit.unit
            )
            if len(distances_out) > 0:
                displacements = estimate_rsd_displacements(
                    classy_output=classy_output,
                    cosmo=self.cosmo,
                    redshifts=z_at_value(self.cosmo.comoving_distance, distances_out),
                    factor=factor,
                )
                if sign < 0:
                    out_indices = distances_out + displacements < lcd_limit
                else:
                    out_indices = distances_out - displacements > lcd_limit
            else:
                out_indices = False
            if np.any(out_indices):
                # If there are displaced coordinates outside the lightcone box, we return the
                # corresponding distance that is closest to the original lightcone boundaries
                lcd_limits_rsd.append(distances_out[out_indices][0])
            else:
                raise ValueError(
                    f"You have set apply_rsds to True with node redshifts between {min(inputs.node_redshifts)} and {max(inputs.node_redshifts)} "
                    f"and lightcone redshifts between {self.lc_redshifts.min()} and {self.lc_redshifts.max()}. "
                    "However, RSDs are expected to contribute the lightcone from higer/lower rdshifts. "
                    "Extend the limits of node redshifts to avoid this error."
                )
        return lcd_limits_rsd

    def __init_subclass__(cls) -> None:
        """Enabe plugin-style behaviour."""
        _LIGHTCONERS[cls.__name__] = cls
        return super().__init_subclass__()


@attrs.define(kw_only=True, slots=False)
class RectilinearLightconer(Lightconer):
    """The class rectilinear lightconer."""

    index_offset: int = attrs.field()
    line_of_sight_axis: int = attrs.field(
        default=-1, converter=int, validator=attrs.validators.in_([-1, 0, 1, 2])
    )

    @index_offset.default
    def _index_offset_default(self) -> int:
        # While it probably makes more sense to use zero as the default offset,
        # we use n_lightcone to maintain default backwards compatibility.
        return len(self.lc_distances)

    def coeval_subselect(
        self, lcd: Quantity[pixel], coeval: np.ndarray, coeval_res: Quantity[_LENGTH]
    ):
        """Sub-select the coeval slice corresponding to this coeval distance."""
        # This makes the back of the lightcone exactly line up with the back of the
        # coeval box at that redshift, modulo the index_offset.
        lcpix = self.get_lc_distances_in_pixels(coeval_res)
        lcidx = int((lcpix.max() - lcd + 1 * pixel).to_value(pixel))
        return coeval.take(-lcidx + self.index_offset, axis=2, mode="wrap")

    def construct_lightcone(
        self,
        lcd: np.ndarray,
        box: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Construct slices of the lightcone between two coevals."""
        return box

    def construct_los_velocity_lightcone(
        self,
        lcd: np.ndarray,
        velocity: np.ndarray,
    ) -> np.ndarray:
        """Construct slices of the lightcone between two coevals."""
        return velocity

    def get_shape(self, simulation_options: SimulationOptions) -> tuple[int, int, int]:
        """Get the shape of the lightcone."""
        return (
            simulation_options.HII_DIM,
            simulation_options.HII_DIM,
            len(self.lc_distances),
        )


def _rotation_eq(x, y):
    """Compare two rotations."""
    if x is None and y is None:
        return True

    return np.allclose(x.as_matrix(), y.as_matrix())


@attrs.define(kw_only=True, slots=False)
class AngularLightconer(Lightconer):
    """Angular lightcone slices constructed from rectlinear coevals."""

    latitude: np.ndarray = attrs.field(eq=attrs.cmp_using(eq=np.allclose))
    longitude: np.ndarray = attrs.field(eq=attrs.cmp_using(eq=np.allclose))
    interpolation_order: int = attrs.field(
        default=1, converter=int, validator=attrs.validators.in_([0, 1, 3, 5])
    )
    origin: Quantity[pixel, (3,), float] = attrs.field(
        eq=attrs.cmp_using(eq=np.allclose)
    )
    rotation: Rotation = attrs.field(
        default=None,
        eq=attrs.cmp_using(eq=_rotation_eq),
        validator=attrs.validators.optional(attrs.validators.instance_of(Rotation)),
    )

    def __attrs_post_init__(self) -> None:
        """Post-init."""
        self._cache = {
            "lcd": None,
            "interpolator": None,
        }

    @longitude.validator
    def _longitude_validator(self, attribute, value):
        if value.ndim != 1:
            raise ValueError("longitude must be 1-dimensional")
        if np.any(value < 0) or np.any(value > 2 * np.pi):
            raise ValueError("longitude must be in the range [0, 2pi]")
        if value.shape != self.latitude.shape:
            raise ValueError("longitude and latitude must have the same shape")

    @origin.default
    def _origin_default(self):
        return np.zeros(3) * pixel

    @classmethod
    def like_rectilinear(
        cls,
        simulation_options: SimulationOptions,
        match_at_z: float,
        cosmo: FLRW = Planck18,
        **kw,
    ):
        """Create an angular lightconer with the same pixel size as a rectilinear one.

        This is useful for comparing the two lightconer types.

        Parameters
        ----------
        simulation_options
            The user parameters.
        match_at_z
            The redshift at which the angular lightconer should match the rectilinear
            one.
        cosmo
            The cosmology to use.

        Other Parameters
        ----------------
        All other parameters passed through to the constructor.

        Returns
        -------
        AngularLightconer
            The angular lightconer.
        """
        box_size_radians = (
            simulation_options.BOX_LEN / cosmo.comoving_distance(match_at_z).value
        )

        lon = np.linspace(0, box_size_radians, simulation_options.HII_DIM)
        # This makes the X-values increasing from 0.
        lat = np.linspace(0, box_size_radians, simulation_options.HII_DIM)[::-1]

        LON, LAT = np.meshgrid(lon, lat)
        LON = LON.flatten()
        LAT = LAT.flatten()

        origin_offset = -cosmo.comoving_distance(match_at_z).to(
            pixel, pixel_scale(simulation_options.cell_size / pixel)
        )
        origin = np.array([0, 0, origin_offset.value]) * origin_offset.unit
        rot = Rotation.from_euler("Y", -np.pi / 2)

        return cls.between_redshifts(
            min_redshift=match_at_z,
            resolution=simulation_options.cell_size,
            latitude=LAT,
            longitude=LON,
            origin=origin,
            rotation=rot,
            **kw,
        )

    def construct_lightcone(
        self,
        lcd: Quantity[pixel],
        box: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Construct the lightcone slices from bracketing coevals."""
        if self._cache["lcd"] == lcd:
            interpolator = self._cache["interpolator"]
        else:
            interpolator = self._refresh_cache(lcd)
        return interpolator(box)

    def construct_los_velocity_lightcone(
        self,
        lcd: Quantity[pixel],
        velocities: Sequence[np.ndarray],
    ):
        """Construct the LoS velocity lightcone from 3D velocities."""
        if self._cache["lcd"] == lcd:
            interpolator = self._cache["interpolator"]
        else:
            interpolator = self._refresh_cache(lcd)
        return next(make_lightcone_slice_vector_field([velocities], interpolator))

    def _refresh_cache(self, lcd):
        result = make_lightcone_slice_interpolator(
            latitude=self.latitude,
            longitude=self.longitude,
            distance_to_shell=lcd,
            interpolation_order=self.interpolation_order,
            origin=self.origin,
            rotation=self.rotation,
        )
        self._cache["lcd"] = lcd
        self._cache["interpolator"] = result
        return result

    def get_shape(self, simulation_options: SimulationOptions) -> tuple[int, int]:
        """Get the shape of the lightcone slices."""
        return (len(self.longitude), len(self.lc_redshifts))

    def validate_options(
        self,
        inputs: InputParameters,
        include_dvdr_in_tau21: bool,
        apply_rsds: bool,
        classy_output: Class | None = None,
    ) -> Lightconer:
        """Validate 21cmFAST options."""
        lightconer = super().validate_options(
            inputs=inputs,
            include_dvdr_in_tau21=include_dvdr_in_tau21,
            apply_rsds=apply_rsds,
            classy_output=classy_output,
        )

        if (
            include_dvdr_in_tau21 or apply_rsds
        ) and not inputs.matter_options.KEEP_3D_VELOCITIES:
            raise ValueError(
                "To account for RSDs or velocity corrections in an angular lightcone, you need to set "
                "matter_options.KEEP_3D_VELOCITIES=True"
            )

        return lightconer



"""Test the C interpolation tables."""

import matplotlib as mpl
import numpy as np
import pytest

from py21cmfast.wrapper import cfuncs as cf

from .produce_integration_test_data import get_all_options_struct, print_failure_stats

# NOTE: The relative tolerance is set to cover the inaccuracy in interpolaton
#       Whereas absolute tolerances are set to avoid issues with minima
#       i.e the SFRD table has a forced minima of exp(-50)
#       Currently (with the #defined bin numbers) the minihalo tables have a ~1.5 % error maximum
#       With >1% error in less than 4% of bins (at the high turnover mass where fcoll is tiny)
#       The rest of the tables (apart from the xfail inverse tables) are well under 1% error

# The tables are much better than 2% but there are some problem bins (very high density, very small mass ranges)
RELATIVE_TOLERANCE = 2e-2

OPTIONS_PS = {
    "EH": [10, {"POWER_SPECTRUM": "EH"}],
    "BBKS": [10, {"POWER_SPECTRUM": "BBKS"}],
    "BE": [10, {"POWER_SPECTRUM": "EFSTATHIOU"}],
    "Peebles": [10, {"POWER_SPECTRUM": "PEEBLES"}],
    "White": [10, {"POWER_SPECTRUM": "WHITE"}],
    "CLASS": [10, {"POWER_SPECTRUM": "CLASS"}],
}

OPTIONS_HMF = {
    "PS": [10, {"HMF": "PS"}],
    "ST": [10, {"HMF": "ST"}],
}

OPTIONS_INTMETHOD = {
    "QAG": "GSL-QAG",
    "GL": "GAUSS-LEGENDRE",
    "FFCOLL": "GAMMA-APPROX",
}

R_PARAM_LIST = [1.5, 5, 10, 30, 60]

options_ps = list(OPTIONS_PS.keys())
options_hmf = list(OPTIONS_HMF.keys())

# This is confusing and we should change the dict to a list
options_intmethod = list(OPTIONS_INTMETHOD.keys())
# the minihalo ffcoll tables have some bins (when Mturn -> M_turn_upper) which go above 10% error compared to their "integrals"
#    they can pass by doubling the number of M_turn bins and setting relative error to 5% but I think this
#    is better left for later


# Test delta range for CMF integrals over cells
@pytest.fixture(scope="module")
def delta_range():
    return np.linspace(-0.98, 1.7, num=800)


# Mass condition range, and integral bound range for testing CMF integrals
@pytest.fixture(scope="module")
def mass_range():
    return np.logspace(7, 13, num=200)


# Mass turnover range for testing CMF/UMF integrals
@pytest.fixture(scope="module")
def log10_mturn_range():
    return np.linspace(5, 8.5, num=40)


# redshift range for testing UMF integrals
@pytest.fixture(scope="module")
def z_range():
    return np.linspace(6, 35, num=100)


def get_delta_subset(inputs, redshift, R, delta_range):
    """Get a selection of delta values away from delta_crit for testing.

    The integrals and tables perform poorly very close to delta_crit.
    This was made a function for uniformity across tests.
    """
    delta_crit = cf.get_delta_crit(
        inputs=inputs,
        mass=cf.get_condition_mass(inputs, R),
        redshift=redshift,
    )
    sel_delta = (delta_crit - delta_range) > 0.1

    return sel_delta


@pytest.mark.parametrize("name", options_ps)
def test_sigma_table(name, mass_range, plt):
    abs_tol = 0
    redshift, kwargs = OPTIONS_PS[name]
    inputs = get_all_options_struct(redshift, **kwargs)["inputs"]

    sigma_tables, dsigma_tables = cf.evaluate_sigma(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        masses=mass_range,
    )

    sigma_integrals, dsigma_integrals = cf.evaluate_sigma(
        inputs=inputs.evolve_input_structs(USE_INTERPOLATION_TABLES="no-interpolation"),
        masses=mass_range,
    )

    if plt == mpl.pyplot:
        make_table_comparison_plot(
            [mass_range, mass_range],
            [None, None],
            [sigma_tables, -dsigma_tables],
            [sigma_integrals, -dsigma_integrals],
            plt,
            reltol=RELATIVE_TOLERANCE,
            xlabels=["Mass", "Mass"],
            ylabels=["sigma", "dsigmasqdM"],
        )

    np.testing.assert_allclose(
        sigma_integrals, sigma_tables, atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )
    np.testing.assert_allclose(
        dsigma_integrals, dsigma_tables, atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )


@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("cond_type", ["cat", "grid"])
def test_massfunc_conditional_tables(name, cond_type, mass_range, delta_range, plt):
    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(redshift, **kwargs)["inputs"]
    from_cat = "cat" in cond_type

    # We limit our delta range here because it seems that the calculations are numerically unstable
    # for high delta values (especially if name = "ST" and cond_type = "grid")
    inputs_cond = mass_range if from_cat else delta_range[delta_range <= 1.44]
    z_desc = (
        (redshift + 1) / inputs.simulation_options.ZPRIME_STEP_FACTOR - 1
        if from_cat
        else None
    )

    # get the tables
    nhalo_tbl, mcoll_tbl = cf.evaluate_condition_integrals(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        redshift=redshift,
        redshift_prev=z_desc,
        cond_array=inputs_cond,
    )
    # get the integrals
    nhalo_exp, mcoll_exp = cf.evaluate_condition_integrals(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        redshift=redshift,
        redshift_prev=z_desc,
        cond_array=inputs_cond,
    )

    if plt == mpl.pyplot:
        make_table_comparison_plot(
            [inputs_cond, inputs_cond],
            [None, None],
            [nhalo_tbl, mcoll_tbl],
            [nhalo_exp, mcoll_exp],
            plt,
            reltol=RELATIVE_TOLERANCE,
            logx=from_cat,
            xlabels=["Mass" if from_cat else "delta"] * 2,
            ylabels=["Nhalo", "Mcoll"],
        )

    print_failure_stats(
        nhalo_tbl,
        nhalo_exp,
        [inputs_cond],
        0.0,
        RELATIVE_TOLERANCE,
        "expected N halo",
    )
    print_failure_stats(
        mcoll_tbl,
        mcoll_exp,
        [inputs_cond],
        0.0,
        RELATIVE_TOLERANCE,
        "expected M halo",
    )

    np.testing.assert_allclose(nhalo_exp, nhalo_tbl, rtol=RELATIVE_TOLERANCE)
    np.testing.assert_allclose(mcoll_exp, mcoll_tbl, rtol=RELATIVE_TOLERANCE)


@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("cond_type", ["cat", "grid"])
def test_inverse_cmf_tables(name, cond_type, delta_range, mass_range, plt):
    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(redshift, **kwargs)["inputs"]

    from_cat = "cat" in cond_type
    lnMmin_range = np.log(mass_range)
    # We limit our delta range here because it seems that the calculations are numerically unstable
    # for high delta values (especially if name = "ST" and cond_type = "grid")
    inputs_cond = mass_range if from_cat else delta_range[delta_range <= 1.44]
    z_desc = (
        (redshift + 1) / inputs.simulation_options.ZPRIME_STEP_FACTOR - 1
        if from_cat
        else None
    )

    # Get the full integrals
    nhalo, _ = cf.evaluate_condition_integrals(
        inputs=inputs,
        redshift=redshift,
        redshift_prev=z_desc,
        cond_array=inputs_cond,
    )
    # Get the Integrals on mass limits
    cmf_integral = cf.integrate_chmf_interval(
        inputs=inputs,
        lnm_lower=lnMmin_range,
        lnm_upper=np.full_like(lnMmin_range, 16 * np.log(10)),
        cond_values=inputs_cond,
        redshift=redshift,
        redshift_prev=z_desc,
    )
    probabilities = cmf_integral / nhalo[:, None]

    inputs_cond_bc, inputs_mmin_bc = np.meshgrid(
        inputs_cond, lnMmin_range, indexing="ij"
    )
    # Take those probabilites to the inverse table
    icmf_table = cf.evaluate_inverse_table(
        inputs=inputs,
        probabilities=probabilities,
        cond_array=inputs_cond_bc,
        redshift=redshift,
        redshift_prev=z_desc,
    )

    # ignore condition out-of-bounds (returned as -1 by evaluate_inverse_table)
    sel_oob = icmf_table < 0.0
    # ignore probability out-of-bounds (extrapolated by the backend)
    sel_lowprob = cmf_integral < np.exp(inputs.simulation_options.MIN_LOGPROB)
    if plt == mpl.pyplot:
        last_cond = np.amax(np.where(np.any(~sel_oob, axis=-1)))
        first_cond = np.amin(np.where(np.any(~sel_oob, axis=-1)))
        sel_plot = np.linspace(first_cond, last_cond, num=6).astype(int)
        make_table_comparison_plot(
            [probabilities[sel_plot, :].T],
            [inputs_cond[sel_plot]],
            [icmf_table[sel_plot, :].T],
            [np.exp(lnMmin_range)],
            plt,
            zlabels=[r"$\delta =$" if not from_cat else r"$M=$"],
            logx=True,
            logy=True,
            label_test=[False, False],
            xlabels=["Probability"],
            ylabels=["Mass"],
            xlim=[np.exp(inputs.simulation_options.MIN_LOGPROB) / 10, 1.0],
            reltol=RELATIVE_TOLERANCE,
        )

    # easiest way to ignore in testing is to fix values to nan
    inputs_mmin_bc[sel_lowprob | sel_oob] = np.nan
    icmf_table[sel_lowprob | sel_oob] = np.nan

    # NOTE: very dense cells can have ~3.5% errors at low mass, other bins much more accurate
    # Previously I didn't test cells near delta_crit, but I think a 5% tolerance test is better
    rtol = 5e-2
    print_failure_stats(
        np.log(icmf_table),
        inputs_mmin_bc,
        [inputs_cond, probabilities],
        0.0,
        rtol,
        "Inverse CMF",
    )

    np.testing.assert_allclose(
        inputs_mmin_bc,
        np.log(icmf_table),
        rtol=rtol,
    )


@pytest.mark.parametrize("R", R_PARAM_LIST)
def test_FgtrM_conditional_tables(R, delta_range, plt):
    redshift, kwargs = OPTIONS_HMF["PS"]  # always erfc
    inputs = get_all_options_struct(redshift, **kwargs)["inputs"]

    fcoll_tables, dfcoll_tables = cf.evaluate_FgtrM_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        redshift=redshift,
        densities=delta_range,
        R=R,
    )
    fcoll_integrals, dfcoll_integrals = cf.evaluate_FgtrM_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        redshift=redshift,
        densities=delta_range,
        R=R,
    )

    abs_tol = 0.0
    if plt == mpl.pyplot:
        make_table_comparison_plot(
            [delta_range, delta_range],
            [None, None],
            [fcoll_tables, np.fabs(dfcoll_tables)],
            [fcoll_integrals, np.fabs(dfcoll_integrals)],
            plt,
            abstol=abs_tol,
            reltol=RELATIVE_TOLERANCE,
            xlabels=["delta", "delta"],
            ylabels=["fcoll", "dfolldz"],
        )

    sel_delta = get_delta_subset(inputs, redshift, R, delta_range)

    print_failure_stats(
        fcoll_tables[sel_delta],
        fcoll_integrals[sel_delta],
        [delta_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "fcoll",
    )

    print_failure_stats(
        dfcoll_tables[sel_delta],
        dfcoll_integrals[sel_delta],
        [delta_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "dfcoll",
    )

    np.testing.assert_allclose(
        fcoll_tables[sel_delta],
        fcoll_integrals[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )
    np.testing.assert_allclose(
        dfcoll_tables[sel_delta],
        dfcoll_integrals[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )


@pytest.mark.parametrize("name", options_hmf)
def test_SFRD_z_tables(name, z_range, log10_mturn_range, plt):
    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(
        redshift,
        USE_MINI_HALOS=True,
        INHOMO_RECO=True,
        USE_TS_FLUCT=True,
        **kwargs,
    )["inputs"]

    z_input, mt_input = np.meshgrid(z_range, log10_mturn_range, indexing="ij")
    SFRD_tables, SFRD_tables_mini = cf.evaluate_SFRD_z(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        redshifts=z_input,
        log10mturns=mt_input,
    )
    SFRD_integrals, SFRD_integrals_mini = cf.evaluate_SFRD_z(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        redshifts=z_input,
        log10mturns=mt_input,
    )

    abs_tol = 1e-5
    if plt == mpl.pyplot:
        xl = log10_mturn_range.size - 1
        sel_m = np.linspace(0, xl, num=5).astype(int)
        make_table_comparison_plot(
            [z_range, z_range],
            [np.array([0]), 10 ** log10_mturn_range[sel_m]],
            [SFRD_tables[..., 0], SFRD_tables_mini[..., sel_m]],
            [SFRD_integrals[..., 0], SFRD_integrals_mini[..., sel_m]],
            plt,
            abstol=abs_tol,
            reltol=RELATIVE_TOLERANCE,
            label_test=[True, False],
            xlabels=["redshift", "redshift"],
            ylabels=["SFRD", "SFRD_mini"],
        )

    print_failure_stats(
        SFRD_tables,
        SFRD_integrals,
        [z_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "SFRD_z",
    )
    print_failure_stats(
        SFRD_tables_mini,
        SFRD_integrals_mini,
        [z_range, 10**log10_mturn_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "SFRD_z_mini",
    )

    np.testing.assert_allclose(
        SFRD_tables, SFRD_integrals, atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )
    np.testing.assert_allclose(
        SFRD_tables_mini, SFRD_integrals_mini, atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )


@pytest.mark.parametrize("name", options_hmf)
def test_Nion_z_tables(name, z_range, log10_mturn_range, plt):
    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(
        redshift,
        USE_MINI_HALOS=True,
        INHOMO_RECO=True,
        USE_TS_FLUCT=True,
        **kwargs,
    )["inputs"]

    z_input, mt_input = np.meshgrid(z_range, log10_mturn_range, indexing="ij")
    nion_tables, nion_tables_mini = cf.evaluate_Nion_z(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        redshifts=z_input,
        log10mturns=mt_input,
    )
    nion_integrals, nion_integrals_mini = cf.evaluate_Nion_z(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        redshifts=z_input,
        log10mturns=mt_input,
    )

    abs_tol = 2e-6
    if plt == mpl.pyplot:
        xl = log10_mturn_range.size - 1
        sel_m = np.linspace(0, xl, num=5).astype(int)
        make_table_comparison_plot(
            [z_range, z_range],
            [np.array([0]), 10 ** log10_mturn_range[sel_m]],
            [nion_tables[..., 0], nion_tables_mini[..., sel_m]],
            [nion_integrals[..., 0], nion_integrals_mini[..., sel_m]],
            plt,
            abstol=abs_tol,
            reltol=RELATIVE_TOLERANCE,
            label_test=[True, False],
            xlabels=["redshift", "redshift"],
            ylabels=["Nion", "Nion_mini"],
        )

    print_failure_stats(
        nion_tables,
        nion_integrals,
        [z_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "SFRD_z",
    )
    print_failure_stats(
        nion_tables_mini,
        nion_integrals_mini,
        [z_range, 10**log10_mturn_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "SFRD_z_mini",
    )

    np.testing.assert_allclose(
        nion_tables, nion_integrals, atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )
    np.testing.assert_allclose(
        nion_tables_mini, nion_integrals_mini, atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )


# A few notes on this test function:
#   Mass limits are set explicitly using the param values
#       and #defines are copied to hard code (not ideal)
#   Density and Mturn limits are set to their maxima since we don't have cubes.
#       Hence this is a worst case scenario
#   While the EvaluateX() functions are useful in the main code to be agnostic to USE_INTERPOLATION_TABLES
#       I do not use them here fully, instead calling the integrals directly to avoid parameter changes
#       Mostly since if we set simulation_options.USE_INTERPOLATION_TABLES=False then the sigma tables aren't used
#       and it takes forever
@pytest.mark.parametrize("mini", ["mini", "acg"])
@pytest.mark.parametrize("R", R_PARAM_LIST)
@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("intmethod", options_intmethod)
def test_Nion_conditional_tables(
    name, log10_mturn_range, delta_range, R, mini, intmethod, plt
):
    if intmethod == "FFCOLL":
        if name != "PS":
            pytest.skip("FAST FFCOLL INTEGRALS WORK ONLY WITH EPS")
        else:
            pytest.xfail(
                "FFCOLL TABLES drop sharply at high Mturn, causing failure at 0.1 levels"
            )

    mini_flag = mini == "mini"
    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(
        redshift,
        USE_MINI_HALOS=mini_flag,
        INHOMO_RECO=True,
        USE_TS_FLUCT=True,
        **kwargs,
    )["inputs"]

    # NOTE: we still pass a 2D array when minihalos are off.
    #   part of the test is that passing different mturns *doesn't* change
    #   the integral without minihalos
    d_input, mt_input = np.meshgrid(delta_range, log10_mturn_range, indexing="ij")

    Nion_tables, Nion_tables_mini = cf.evaluate_Nion_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        redshift=redshift,
        radius=R,
        densities=d_input,
        l10mturns_mcg=mt_input,
        l10mturns_acg=mt_input,
    )

    Nion_integrals, Nion_integrals_mini = cf.evaluate_Nion_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        redshift=redshift,
        radius=R,
        densities=d_input,
        l10mturns_mcg=mt_input,
        l10mturns_acg=mt_input,
    )

    # The bilinear interpolation we use underperforms at high mturn due to the sharp
    # dropoff. Setting an absolute tolerance to a level we care about for reionisation
    # rather than a level we expect from interp tables remedies this for now.
    # TODO: In future we should investigate cubic splines etc.
    abs_tol = 1e-8

    if plt == mpl.pyplot:
        xl = log10_mturn_range.size - 1
        sel_m = np.linspace(0, xl, num=5).astype(int)
        Nion_tb_plot = Nion_tables[..., sel_m]
        Nion_il_plot = Nion_integrals[..., sel_m]

        make_table_comparison_plot(
            [delta_range, delta_range],
            [np.array([0]), 10 ** log10_mturn_range[sel_m]],
            [Nion_tb_plot[..., 0], Nion_tables_mini[..., sel_m]],
            [Nion_il_plot[..., 0], Nion_integrals_mini[..., sel_m]],
            plt,
            abstol=abs_tol,
            reltol=RELATIVE_TOLERANCE,
            label_test=[True, False],
            xlabels=["delta", "delta"],
            ylabels=["Nion", "Nion_mini"],
        )

    sel_delta = get_delta_subset(inputs, redshift, R, delta_range)

    print_failure_stats(
        Nion_tables[sel_delta],
        Nion_integrals[sel_delta],
        [delta_range[sel_delta], 10**log10_mturn_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "Nion_c",
    )

    if mini_flag:
        print_failure_stats(
            Nion_tables_mini[sel_delta],
            Nion_integrals_mini[sel_delta],
            [delta_range[sel_delta], 10**log10_mturn_range],
            abs_tol,
            RELATIVE_TOLERANCE,
            "Nion_c_mini",
        )

    np.testing.assert_allclose(
        Nion_tables[sel_delta],
        Nion_integrals[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )

    np.testing.assert_allclose(
        Nion_tables_mini[sel_delta],
        Nion_integrals_mini[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )


@pytest.mark.parametrize("mini", ["mini", "acg"])
@pytest.mark.parametrize("R", R_PARAM_LIST)
@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("intmethod", options_intmethod)
def test_Xray_conditional_tables(
    name, log10_mturn_range, delta_range, R, mini, intmethod, plt
):
    if intmethod == "FFCOLL":
        if name != "PS":
            pytest.skip("FAST FFCOLL INTEGRALS WORK ONLY WITH EPS")
        else:
            pytest.xfail(
                "FFCOLL TABLES drop sharply at high Mturn, causing failure at 0.1 levels"
            )

    mini_flag = mini == "mini"

    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(
        redshift,
        USE_MINI_HALOS=mini_flag,
        INHOMO_RECO=True,
        USE_TS_FLUCT=True,
        **kwargs,
    )["inputs"]

    d_input, mt_input = np.meshgrid(delta_range, log10_mturn_range, indexing="ij")

    Xray_tables = cf.evaluate_Xray_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        redshift=redshift,
        radius=R,
        densities=d_input,
        log10mturns=mt_input,
    )

    Xray_integrals = cf.evaluate_Xray_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        redshift=redshift,
        radius=R,
        densities=d_input,
        log10mturns=mt_input,
    )

    abs_tol = 0.0
    if plt == mpl.pyplot:
        xl = log10_mturn_range.size - 1
        sel_m = np.linspace(0, xl, num=5).astype(int)
        Xray_tb_plot = Xray_tables[..., sel_m]
        Xray_il_plot = Xray_integrals[..., sel_m]
        make_table_comparison_plot(
            [delta_range],
            [10 ** log10_mturn_range[sel_m]],
            [Xray_tb_plot],
            [Xray_il_plot],
            plt,
            abstol=abs_tol,
            reltol=RELATIVE_TOLERANCE,
            label_test=[True],
            xlabels=["delta"],
            ylabels=["Lx"],
        )

    sel_delta = get_delta_subset(inputs, redshift, R, delta_range)

    print_failure_stats(
        Xray_tables[sel_delta],
        Xray_integrals[sel_delta],
        [delta_range[sel_delta], 10**log10_mturn_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "Xray_c",
    )

    np.testing.assert_allclose(
        Xray_tables[sel_delta],
        Xray_integrals[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )


@pytest.mark.parametrize("R", R_PARAM_LIST)
@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("intmethod", options_intmethod)
def test_SFRD_conditional_table(
    name, log10_mturn_range, delta_range, R, intmethod, plt
):
    if intmethod == "FFCOLL":
        if name != "PS":
            pytest.skip("FAST FFCOLL INTEGRALS WORK ONLY WITH EPS")
        else:
            pytest.xfail("FFCOLL TABLES drop sharply at high Mturn, causing failure")

    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(
        redshift,
        USE_MINI_HALOS=True,
        INHOMO_RECO=True,
        USE_TS_FLUCT=True,
        **kwargs,
    )["inputs"]

    d_input, mt_input = np.meshgrid(delta_range, log10_mturn_range, indexing="ij")

    SFRD_tables, SFRD_tables_mini = cf.evaluate_SFRD_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="hmf-interpolation"
        ),
        radius=R,
        redshift=redshift,
        densities=d_input,
        log10mturns=mt_input,
    )

    SFRD_integrals, SFRD_integrals_mini = cf.evaluate_SFRD_cond(
        inputs=inputs.evolve_input_structs(
            USE_INTERPOLATION_TABLES="sigma-interpolation"
        ),
        radius=R,
        redshift=redshift,
        densities=d_input,
        log10mturns=mt_input,
    )

    # The bilinear interpolation we use underperforms at high mturn due to the sharp
    # dropoff. Setting an absolute tolerance to a level we care about for reionisation
    # rather than a level we expect from interp tables remedies this for now.
    # TODO: In future we should investigate cubic splines etc.
    abs_tol = 1e-8
    if plt == mpl.pyplot:
        xl = log10_mturn_range.size - 1
        sel_m = np.linspace(0, xl, num=5).astype(int)
        make_table_comparison_plot(
            [delta_range, delta_range],
            [np.array([0]), 10 ** log10_mturn_range[sel_m]],
            [SFRD_tables[:, 0], SFRD_tables_mini[..., sel_m]],
            [SFRD_integrals[:, 0], SFRD_integrals_mini[..., sel_m]],
            plt,
            abstol=abs_tol,
            reltol=RELATIVE_TOLERANCE,
            label_test=[True, False],
            xlabels=["delta", "delta"],
            ylabels=["SFRD", "SFRD_mini"],
        )

    sel_delta = get_delta_subset(inputs, redshift, R, delta_range)

    print_failure_stats(
        SFRD_tables[sel_delta],
        SFRD_integrals[sel_delta],
        [delta_range[sel_delta]],
        abs_tol,
        RELATIVE_TOLERANCE,
        "SFRD_c",
    )
    print_failure_stats(
        SFRD_tables_mini[sel_delta],
        SFRD_integrals_mini[sel_delta],
        [delta_range[sel_delta], 10**log10_mturn_range],
        abs_tol,
        RELATIVE_TOLERANCE,
        "SFRD_c_mini",
    )

    np.testing.assert_allclose(
        SFRD_tables[sel_delta],
        SFRD_integrals[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )
    np.testing.assert_allclose(
        SFRD_tables_mini[sel_delta],
        SFRD_integrals_mini[sel_delta],
        atol=abs_tol,
        rtol=RELATIVE_TOLERANCE,
    )


INTEGRAND_OPTIONS = ["sfrd", "n_ion"]


@pytest.mark.parametrize("R", R_PARAM_LIST)
@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("integrand", INTEGRAND_OPTIONS)
def test_conditional_integral_methods(
    R, log10_mturn_range, delta_range, name, integrand, plt
):
    redshift, kwargs = OPTIONS_HMF[name]
    inputs = get_all_options_struct(
        redshift,
        USE_MINI_HALOS=True,
        INHOMO_RECO=True,
        USE_TS_FLUCT=True,
        **kwargs,
    )["inputs"]

    d_input, mt_input = np.meshgrid(delta_range, log10_mturn_range, indexing="ij")

    integrals = []
    integrals_mini = []
    for method in ["GSL-QAG", "GAUSS-LEGENDRE", "GAMMA-APPROX"]:
        print(f"Starting method {method}", flush=True)
        if name != "PS" and method == "GAMMA-APPROX":
            continue

        inputs = inputs.evolve_input_structs(
            INTEGRATION_METHOD_ATOMIC=method,
            INTEGRATION_METHOD_MINI=method,
        )

        if "sfr" in integrand:
            buf, buf_mini = cf.evaluate_SFRD_cond(
                inputs=inputs,
                redshift=redshift,
                radius=R,
                densities=d_input,
                log10mturns=mt_input,
            )
        else:
            buf, buf_mini = cf.evaluate_Nion_cond(
                inputs=inputs,
                redshift=redshift,
                radius=R,
                densities=d_input,
                l10mturns_acg=mt_input,
                l10mturns_mcg=mt_input,
            )
        integrals.append(buf)
        integrals_mini.append(buf_mini)

    abs_tol = 1e-6  # minimum = exp(-40) ~1e-18
    if plt == mpl.pyplot:
        xl = log10_mturn_range.size - 1
        sel_m = np.linspace(0, xl, num=5).astype(int)
        iplot = [i[..., sel_m] if i.ndim == 2 else i for i in integrals]
        iplot_mini = [i[..., sel_m] for i in integrals_mini]
        make_integral_comparison_plot(
            delta_range,
            10 ** log10_mturn_range[sel_m],
            iplot,
            iplot_mini,
            plt,
        )

    np.testing.assert_allclose(
        integrals[1], integrals[0], atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )
    np.testing.assert_allclose(
        integrals_mini[1], integrals_mini[0], atol=abs_tol, rtol=RELATIVE_TOLERANCE
    )


def make_table_comparison_plot(
    x,
    tb_z,
    tables,
    integrals,
    plt,
    **kwargs,
):
    # rows = values,fracitonal diff, cols = 1d table, 2d table
    _fig, axs = plt.subplots(
        nrows=2, ncols=len(x), figsize=(12 * len(x) / 2, 9), squeeze=False
    )
    xlabels = kwargs.pop("xlabels", ["delta"] * len(x))
    ylabels = kwargs.pop("ylabels", ["MF_integral"] * len(x))
    zlabels = kwargs.pop("zlabels", ["Mturn"] * len(x))
    label_flags = kwargs.pop("label_test", [True] * len(tb_z))
    for j, z in enumerate(tb_z):
        n_lines = z.size if z is not None else 1
        for i in range(n_lines):
            zlab = zlabels[j] + f" = {z[i]:.2e}" if z is not None else ""
            # allow single arrays
            x_plot = x[j][:, i] if len(x[j].shape) > 1 else x[j]
            i_plot = integrals[j][:, i] if len(integrals[j].shape) > 1 else integrals[j]
            t_plot = tables[j][:, i] if len(tables[j].shape) > 1 else tables[j]
            make_comparison_plot(
                x_plot,
                i_plot,
                t_plot,
                ax=axs[:, j],
                xlab=xlabels[j],
                ylab=ylabels[j],
                label_base=zlab,
                label_test=label_flags[j],
                logx=kwargs.get("logx", False),
                xlim=kwargs.get("xlim"),
                reltol=kwargs.get("reltol"),
                color=f"C{i:d}",
            )


# slightly different from comparison plot since each integral shares a "truth"
def make_integral_comparison_plot(x1, x2, integral_list, integral_list_second, plt):
    _fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(16, 8))

    styles = ["-", ":", "--"]
    for i, (i_first, i_second) in enumerate(
        zip(integral_list, integral_list_second, strict=True)
    ):
        comparison = integral_list[0]
        if len(i_first.shape) == 1:
            i_first = i_first[:, None]
            comparison = integral_list[0][:, None]
        for j in range(i_first.shape[1]):
            axs[0, 0].semilogy(
                x1, i_first[:, j], color=f"C{j:d}", linestyle=styles[i], linewidth=2
            )
            axs[1, 0].semilogy(
                x1,
                i_first[:, j] / comparison[:, j],
                color=f"C{j:d}",
                linestyle=styles[i],
                linewidth=2,
            )

        for j in range(i_second.shape[1]):
            axs[0, 1].semilogy(x1, i_second[:, j], color=f"C{j:d}", linestyle=styles[i])
            axs[1, 1].semilogy(
                x1,
                i_second[:, j] / integral_list_second[0][:, j],
                color=f"C{j:d}",
                linestyle=styles[i],
            )

    axs[1, 0].set_xlabel("delta")
    axs[1, 1].set_xlabel("delta")
    axs[1, 0].set_ylabel("Integral")
    axs[0, 0].set_ylabel("Integral")
    axs[0, 0].grid()
    axs[0, 1].grid()
    axs[1, 0].grid()
    axs[1, 1].grid()


# copied and expanded from test_integration_features.py
def make_comparison_plot(
    x,
    true,
    test,
    ax,
    logx=True,
    logy=True,
    xlab=None,
    ylab=None,
    xlim=None,
    abstol=None,
    reltol=None,
    label_base="",
    label_test=True,
    **kwargs,
):
    true_label = label_base + " True" if label_test else label_base
    test_label = label_base + " Test" if label_test else None
    if logx:
        ax[0].set_xscale("log")
        ax[1].set_xscale("log")
    if logy:
        ax[0].set_yscale("log")
    if xlab:
        ax[0].set_xlabel(xlab)
        ax[1].set_xlabel(xlab)
    if ylab:
        ax[0].set_ylabel(ylab)
    if xlim:
        ax[0].set_xlim(xlim)
        ax[1].set_xlim(xlim)

    ax[0].grid()
    ax[1].grid()

    ax[0].plot(x, true, label=true_label, linestyle="-", **kwargs)
    ax[0].plot(x, test, label=test_label, linestyle=":", linewidth=3, **kwargs)
    ax[0].legend()

    ax[1].plot(x, (test - true) / true, **kwargs)
    ax[1].set_ylabel("Fractional Difference")
    if reltol:
        ax[1].set_ylim([-2 * reltol, 2 * reltol])
        ax[1].axhline(reltol, color="gray", linestyle="--")
        ax[1].axhline(-reltol, color="gray", linestyle="--")



"""Tests of the classy module."""

import numpy as np
import pytest
from astropy import units

from py21cmfast.wrapper.classy_interface import (
    compute_rms,
    run_classy,
)


def test_classy_runs():
    """Test classy runs in non trivial configurations."""
    with pytest.raises(
        KeyError,
        match="You specified m_ncdm, but set N_ncdm=0",
    ):
        classy_output = run_classy(
            N_ncdm=0,
            m_ncdm="0.06",
        )

    classy_output = run_classy(
        N_ncdm=0,
        P_k_max=1.0 / units.Mpc,
    )
    assert classy_output.Om_ncdm(0) == 0.0
    assert max(classy_output.get_transfer(z=0)["k (h/Mpc)"] * classy_output.h()) < 10.0


def test_classy_runs_with_sigma8_A_s(default_input_struct):
    """Test classy runs with sigma and A_s."""
    A_s = 3.0e-9
    sigma8 = 1.0

    classy_output = run_classy()
    assert np.isclose(
        classy_output.sigma8(), default_input_struct.cosmo_params.SIGMA_8
    )  # Not sure why they are not exactly the same
    assert np.isclose(
        classy_output.get_current_derived_parameters(["A_s"])["A_s"],
        default_input_struct.cosmo_params.A_s,
    )

    classy_output = run_classy(sigma8=sigma8)
    assert np.isclose(
        classy_output.sigma8(), sigma8
    )  # Not sure why they are not exactly the same

    classy_output = run_classy(A_s=A_s)
    assert classy_output.get_current_derived_parameters(["A_s"])["A_s"] == A_s

    with pytest.raises(
        KeyError,
        match=r"Do not provide both 'sigma8' and 'A_s' as arguments. Only one of them is allowed.",
    ):
        classy_output = run_classy(sigma8=sigma8, A_s=A_s)


def test_compute_rms(default_input_struct):
    """Test the rms function of classy."""
    # Check that CLASS returns sigma8, and that rms of v_cb at recombination is between 25 to 30 km/s
    classy_output = run_classy()
    rms_d = compute_rms(
        classy_output, kind="d_m", redshifts=0, smoothing_radius=8.0 / classy_output.h()
    )
    rms_vcb = compute_rms(
        classy_output, kind="v_cb", redshifts=1020, smoothing_radius=0
    )
    assert np.isclose(rms_d, default_input_struct.cosmo_params.SIGMA_8)
    assert np.isclose(rms_d, classy_output.sigma8())
    assert rms_vcb < 30.0 * units.km / units.s
    assert rms_vcb > 25.0 * units.km / units.s

    # Same tests as above, but with synchronous gauge
    classy_output = run_classy(gauge="synchronous")
    rms_d = compute_rms(
        classy_output,
        kind="d_m",
        redshifts=0,
        smoothing_radius=8.0 / classy_output.h() * units.Mpc,
    )
    rms_vcb = compute_rms(
        classy_output, kind="v_cb", redshifts=1020, smoothing_radius=0
    )
    assert np.isclose(rms_d, default_input_struct.cosmo_params.SIGMA_8)
    assert np.isclose(rms_d, classy_output.sigma8())
    assert rms_vcb < 30.0 * units.km / units.s
    assert rms_vcb > 25.0 * units.km / units.s


def test_compute_rms_bad_inputs():
    """Test the rms function of classy with bad inputs."""
    classy_output = run_classy()

    with pytest.raises(
        ValueError, match="'kind' can only be d_b, d_cdm, d_m, v_b, v_cdm or v_cb"
    ):
        compute_rms(
            classy_output,
            kind="X",
            redshifts=0,
            smoothing_radius=8.0 / classy_output.h() * units.Mpc,
        )

    with pytest.raises(
        ValueError, match="The units of R_smooth are not of type length!"
    ):
        compute_rms(
            classy_output,
            kind="d_m",
            redshifts=0,
            smoothing_radius=8.0 / classy_output.h() * units.s,
        )



"""Module to deal with the cache.

The module has a manager that essentially establishes a database of cached files,
and provides methods to handle the caching of output data (i.e. determining the
filename for a given set of parameters).
"""

import logging
import re
from hashlib import md5
from pathlib import Path
from typing import ClassVar, Self

import attrs
import numpy as np

from .._cfg import config
from ..wrapper import outputs as op
from ..wrapper.inputs import InputParameters
from ..wrapper.outputs import (
    _ALL_OUTPUT_STRUCTS,
    OutputStruct,
    _HashType,
)
from .h5 import read_inputs, read_output_struct, write_output_to_hdf5

logger = logging.getLogger(__name__)


@attrs.define(frozen=True)
class OutputCache:
    """An object that manages cache files from 21cmFAST simulations.

    This object has a single attribute -- the top-level directory of the cache. This
    directory can be anywhere on disk. A number of methods exist on the object to
    interact with the cache, including finding existing cache files for a particular
    OutputStruct, writing/reading an OutputStruct to/from the cache, and listing
    existing datasets.

    The cache is meant for single-field OutputStruct objects, not "collections" of
    outputs in an evolved universe (like Coeval or Lightcone objects).
    """

    direc: Path = attrs.field(
        default=Path(config["direc"]).expanduser(), converter=Path
    )

    _output_to_cache_map: ClassVar = {
        kls.__name__: kls._compat_hash
        for kls in _ALL_OUTPUT_STRUCTS.values()
        if not kls._meta
    }
    _path_structures: ClassVar = {
        _HashType.user_cosmo: "{matter_cosmo}/{seed}/InitialConditions.h5",
        _HashType.zgrid: "{matter_cosmo}/{seed}/{zgrid}/{redshift}/{cls}.h5",
        _HashType.full: "{matter_cosmo}/{seed}/{zgrid}/{redshift}/{astro_flag}/{cls}.h5",
    }

    @classmethod
    def _get_hashes(cls, inputs: InputParameters) -> dict[str, str]:
        """Return a dict of hashes for different components of the calculation."""
        # Python builtin hashes can be negative which looks weird in filenames
        return {
            "matter_cosmo": md5(
                (
                    repr(inputs.cosmo_params)
                    + repr(inputs.simulation_options)
                    + repr(inputs.matter_options)
                ).encode()
            ).hexdigest(),
            "seed": inputs.random_seed,
            "zgrid": md5(repr(inputs.node_redshifts).encode()).hexdigest(),
            "astro_flag": md5(
                (repr(inputs.astro_params) + repr(inputs.astro_options)).encode()
            ).hexdigest(),
        }

    @classmethod
    def _fill_path_template(
        cls,
        *,
        kind: str | None = None,
        inputs: InputParameters | None = None,
        all_seeds: bool = False,
        redshift: float | None = None,
    ) -> str:
        """Fill the path templates with given values.

        Does the conditional formatting required for each field,
        since we don't do number formatting for the wildcards.

        Returns the filled template path with optional wildcards for
        searching.
        """
        # get the hashes
        if inputs is not None:
            hashes = cls._get_hashes(inputs)
            # format required hashes to string
            hashes["seed"] = r"\d+" if all_seeds else f"{hashes['seed']:d}"
            hashes["matter_cosmo"] = f"{hashes['matter_cosmo']}"
            hashes["zgrid"] = f"{hashes['zgrid']}"
            hashes["astro_flag"] = f"{hashes['astro_flag']}"
        else:
            hashes = {
                "matter_cosmo": ".+?",
                "seed": r"\d+",
                "zgrid": ".+?",
                "astro_flag": ".+?",
            }

        # do the conditional formatting
        hashes["redshift"] = f"{redshift:.4f}" if redshift is not None else ".+?"
        hashes["cls"] = kind if kind in cls._output_to_cache_map else ".+?"

        # precedence: outputclass mapped (class name -> template) > template provided as _HashType (template directly) > full astro path
        path_template = cls._output_to_cache_map.get(kind, kind)
        template = cls._path_structures.get(
            path_template, cls._path_structures[_HashType.full]
        )
        template = template.format(**hashes)
        return template

    def get_filename(self, obj: OutputStruct) -> str:
        """
        Generate a filename for a given OutputStruct object based on its properties.

        This method constructs a unique filename using the object's class name, redshift
        (if available), and hashes of its input parameters. The filename structure is
        determined by the _path_structures dictionary.

        Parameters
        ----------
        obj : OutputStruct
            The OutputStruct object for which to generate a filename.

        Returns
        -------
        str
            The generated filename for the given OutputStruct object.
        """
        return self._fill_path_template(
            kind=obj.__class__.__name__,
            redshift=getattr(obj, "redshift", None),
            inputs=obj.inputs,
            all_seeds=False,
        )

    def get_path(self, obj: OutputStruct) -> Path:
        """
        Get the full path for a given OutputStruct object.

        This method combines the cache directory with the filename generated
        for the given OutputStruct object to create a complete file path.

        Parameters
        ----------
        obj : OutputStruct
            The OutputStruct object for which to generate the full path.

        Returns
        -------
        Path
            The complete file path for the given OutputStruct object.
        """
        return self.direc / self.get_filename(obj)

    def find_existing(self, obj: OutputStruct) -> Path | None:
        """
        Try to find existing boxes which match the parameters of this instance.

        Parameters
        ----------
        obj : OutputStruct
            The OutputStruct instance to search for.

        Returns
        -------
        Path
            The path to an existing cached OutputStruct matching this instance, or
            None if no match is found.
        """
        # Try an explicit path
        f = self.get_path(obj)
        return f if f.exists() else None

    def write(self, obj: OutputStruct) -> None:
        """
        Write an OutputStruct object to the cache.

        This method writes the given OutputStruct object to an HDF5 file in the cache,
        using the path determined by the object's properties.

        Parameters
        ----------
        obj : OutputStruct
            The OutputStruct object to be written to the cache.
        """
        pth = self.get_path(obj)
        write_output_to_hdf5(obj, path=pth)

    def list_datasets(
        self,
        *,
        kind: str | None = None,
        inputs: InputParameters | None = None,
        all_seeds: bool = True,
        redshift: float | None = None,
    ) -> list[Path]:
        """Return all datasets in the cache which match a given set of filters.

        Parameters
        ----------
        kind: str, optional
            Filter by this kind (a class name of an OutputStruct).
        inputs : InputParameters
            Filter by these input parameters
        all_seeds
            Set to False to only include the seed within `inputs`.
        redshift
            The redshift to search for.

        Returns
        -------
        files
            list of paths pointing to files matching the filters.
        """
        kinds_list = self._output_to_cache_map.keys() if kind is None else [kind]
        templates = [
            self._fill_path_template(
                kind=k,
                inputs=inputs,
                all_seeds=all_seeds,
                redshift=redshift,
            )
            for k in kinds_list
        ]

        allfiles = self.direc.glob("**/*")
        matches = []
        for fl in allfiles:
            for template in templates:
                match = re.search(template, str(fl))
                if match is not None:
                    matches.append(Path(match.string))

        return matches

    def load(self, obj: OutputStruct) -> OutputStruct:
        """Load a cache-backed object from disk corresponding to a given object."""
        existing = self.find_existing(obj)

        if existing is None:
            raise OSError(f"No cache exists for {obj} yet!")

        return read_output_struct(existing, struct=obj.__class__.__name__)


def _pathfield():
    return attrs.field(
        default=None,
        converter=attrs.converters.optional(Path),
    )


def _dict_of_paths_field():
    def _convert(x: dict | None) -> tuple[Path]:
        if x is None:
            return x

        if isinstance(x, dict):
            return {float(z): Path(d) for z, d in x.items()}

    return attrs.field(
        default=None,
        converter=_convert,
    )


@attrs.define
class RunCache:
    """An object that specifies all cache files that should/can exist for a full run.

    This object should be instantiated via the `.from_inputs()` class method.

    The instance simply holds references to all possible cache files for a particular
    total simulation (including all evolution over redshift). Not all of these files
    might exist: if a file doesn't exist it implies that the simulation has not run
    for that redshift/field yet. Attributes with values of None are not meant to exist
    as part of the simulation (e.g. they may be TsBox instances when USE_TS_FLUCT=False).
    """

    InitialConditions: Path = _pathfield()
    PerturbedField: dict[float, Path] = _dict_of_paths_field()
    TsBox: dict[float, Path] = _dict_of_paths_field()
    IonizedBox: dict[float, Path] = _dict_of_paths_field()
    BrightnessTemp: dict[float, Path] = _dict_of_paths_field()
    HaloBox: dict[float, Path] | None = _dict_of_paths_field()
    HaloCatalog: dict[float, Path] | None = _dict_of_paths_field()
    XraySourceBox: dict[float, Path] | None = _dict_of_paths_field()
    inputs: InputParameters | None = attrs.field(default=None)

    @classmethod
    def from_inputs(cls, inputs: InputParameters, cache: OutputCache) -> Self:
        """
        Create a RunCache instance from input parameters and an OutputCache.

        This method generates file paths for various output structures based on the
        provided input parameters and cache configuration.

        Parameters
        ----------
        inputs : InputParameters
            The input parameters for the simulation.
        cache : OutputCache
            The output cache object containing directory and path structure information.

        Returns
        -------
        RunCache
            A new RunCache instance with file paths for various output structures.
        """
        ics = cache.direc / cache._fill_path_template(
            kind="InitialConditions",
            inputs=inputs,
        )

        others = {
            "PerturbedField": {},
            "IonizedBox": {},
            "BrightnessTemp": {},
        }
        if inputs.astro_options.USE_TS_FLUCT:
            others |= {"TsBox": {}}
        if inputs.matter_options.lagrangian_source_grid:
            others |= {"XraySourceBox": {}, "HaloBox": {}}
        if inputs.matter_options.has_discrete_halos:
            others |= {"HaloCatalog": {}}

        for z in inputs.node_redshifts:
            for name, val in others.items():
                val[z] = cache.direc / cache._fill_path_template(
                    kind=name,
                    redshift=z,
                    inputs=inputs,
                )

        return cls(
            InitialConditions=ics,
            **others,
            inputs=inputs,
        )

    @classmethod
    def from_example_file(cls, path: Path | str) -> Self:
        """Create a RunCache object from an example file.

        This method can be used to determine all the cache files that make up a full
        simulation, given a single example file. Note that this method is somewhat
        ambiguous when the input file is "high up" in the simulation heirarchy (e.g.
        InitialConditions or PerturbedField) because the input parameters to these
        objects may differ from those of the full simulation, in their astro_params
        and astro_options. For this reason, it is better to supply a cache object like
        IonizedBox or BrightnessTemp.

        Parameters
        ----------
        path : Path | str
            The path to a particular file in cache. The returned OutputCache object
            will include this file.
        """
        inputs = read_inputs(Path(path))

        for kind in OutputCache._output_to_cache_map:
            template = OutputCache._fill_path_template(
                kind=kind,
                redshift=None,
                inputs=inputs,
                all_seeds=False,
            )
            match = re.search(template, str(path))
            if match is not None:
                parent = Path(str(path)[: match.start()])
                break
        else:
            raise ValueError(
                f"The file {path} does not seem to be within a cache structure."
            )

        return cls.from_inputs(inputs, OutputCache(parent))

    def is_complete_at(
        self, z: float | None = None, index: float | None = None
    ) -> bool:
        """Determine whether the simulation has been completed down to a given redshift."""
        if index is not None and z is not None:
            raise ValueError("Cannot specify both z and index")
        if index is not None:
            z = self.inputs.node_redshifts[index]

        for kind in attrs.asdict(self, recurse=False).values():
            if not isinstance(kind, dict):
                continue
            if not kind[z].exists():
                return False
        return True

    def get_output_struct_at_z(
        self,
        kind: type[OutputStruct] | str,
        z: float | None = None,
        index: int | None = None,
        match_z_within: float = 0.01,
    ):
        """Return an output struct of a given kind at or close to a given redshift.

        Parameters
        ----------
        z : float
            The redshift at which to return an output struct.
        index : int
            The node-redshift index at which to return the output struct.
        allow_closest : bool
            Whether to allow the closest redshift available in the cache to be returned.

        Returns
        -------
        OutputStruct
            The output struct corresponding to the kind and redshift.
        """
        if not isinstance(kind, str):
            kind = kind.__name__
        if kind not in attrs.fields_dict(self.__class__):
            raise ValueError(f"Unknown output kind: {kind}")
        if index is not None:
            if z is not None:
                raise ValueError("Cannot specify both z and index")
            z = self.inputs.node_redshifts[index]

        zs_of_kind = np.array(list(getattr(self, kind).keys()))
        if z not in zs_of_kind:
            closest = np.argmin(np.abs(zs_of_kind - z))
            if abs(zs_of_kind[closest] - z) > match_z_within:
                raise ValueError(
                    f"No output struct found for kind '{kind}' at redshift {z} (closest available: {zs_of_kind[closest]} at z idx = {closest})"
                )
            z = zs_of_kind[closest]

        fl = getattr(self, kind)[z]
        return read_output_struct(fl)

    def get_ics(self) -> op.InitialConditions:
        """Return the initial conditions."""
        return read_output_struct(self.InitialConditions)

    def get_all_boxes_at_z(
        self,
        z: float | None = None,
        index: int | None = None,
        match_z_within: float = 0.01,
        return_ics: bool = False,
    ) -> dict[str, OutputStruct]:
        """Return all boxes at or close to a given redshift.

        Parameters
        ----------
        z : float
            The redshift at which to return the boxes.
        index : int
            The node-redshift index at which to return the boxes.
        match_z_within : float
            The maximum difference between the requested and closest available redshift.

        Returns
        -------
        dict[str, Box]
            A dictionary mapping box names to their corresponding Box instances.
        """
        kinds = [
            k
            for k, v in attrs.asdict(self, recurse=False).items()
            if isinstance(v, dict)
        ]

        out = {
            k: self.get_output_struct_at_z(k, z, index, match_z_within) for k in kinds
        }
        if return_ics:
            out["InitialConditions"] = self.get_ics()
        return out

    def get_coeval_at_z(
        self,
        z: float | None = None,
        index: int | None = None,
        match_z_within: float = 0.01,
    ):
        """Return a Coeval object at or close to a given redshift.

        Parameters
        ----------
        z : float
            The redshift at which to return the Coeval object.
        index : int
            The node-redshift index at which to return the Coeval object.
        match_z_within : float
            The maximum difference between the requested and closest available redshift.

        Returns
        -------
        Coeval
            The Coeval object at the given redshift.
        """
        from py21cmfast.drivers.coeval import Coeval

        boxes = self.get_all_boxes_at_z(z, index, match_z_within, return_ics=True)
        return Coeval(
            initial_conditions=boxes["InitialConditions"],
            perturbed_field=boxes["PerturbedField"],
            ionized_box=boxes["IonizedBox"],
            brightness_temperature=boxes["BrightnessTemp"],
            ts_box=boxes.get("TsBox"),
            halobox=boxes.get("HaloBox"),
        )

    def is_complete(self) -> bool:
        """Whether the cache for the full simulation is complete."""
        if not self.InitialConditions.exists():
            return False

        for kind in attrs.asdict(self, recurse=False).values():
            if not isinstance(kind, dict):
                continue

            for fl in kind.values():
                if not fl.exists():
                    return False
        return True


@attrs.define
class CacheConfig:
    """A configuration object that specifies whether a certain field should be cached."""

    initial_conditions: bool = attrs.field(default=True, converter=bool)
    perturbed_field: bool = attrs.field(default=True, converter=bool)
    spin_temp: bool = attrs.field(default=True, converter=bool)
    ionized_box: bool = attrs.field(default=True, converter=bool)
    brightness_temp: bool = attrs.field(default=True, converter=bool)
    halobox: bool = attrs.field(default=True, converter=bool)
    halo_catalog: bool = attrs.field(default=True, converter=bool)
    xray_source_box: bool = attrs.field(default=True, converter=bool)

    @classmethod
    def on(cls) -> Self:
        """Generate a CacheConfig where all boxes are cached."""
        return cls()

    @classmethod
    def off(cls):
        """Generate a CacheConfig where no boxes are cached."""
        return cls(
            initial_conditions=False,
            perturbed_field=False,
            spin_temp=False,
            ionized_box=False,
            brightness_temp=False,
            halobox=False,
            halo_catalog=False,
            xray_source_box=False,
        )

    @classmethod
    def noloop(cls):
        """Generate a CacheConfig where only boxes not requiring evolution are cached."""
        return cls(
            initial_conditions=True,
            perturbed_field=True,
            spin_temp=False,
            ionized_box=False,
            brightness_temp=False,
            halobox=False,
            halo_catalog=True,
            xray_source_box=False,
        )

    @classmethod
    def last_step_only(cls):
        """Generate a CacheConfig where only boxes needed from more than one step away are cached.

        This represents the minimum caching setup which will *never* store every redshift in memory.
        PerturbedField and PerturbedHaloCatalogs are all calculated at the start of the run, and HaloBox
        is required at multiple redshifts for the XraySourceBox. So this caching setup allows free
        purging of these objects without losing data.
        """
        return cls(
            initial_conditions=False,
            perturbed_field=True,
            spin_temp=False,
            ionized_box=False,
            brightness_temp=False,
            halobox=True,
            halo_catalog=True,
            xray_source_box=False,
        )



"""Utilities for interacting with 21cmFAST data structures."""

from operator import eq
from typing import Any

from .wrapper.inputs import InputParameters

# allow print functions and ambiguous unicode characters
# ruff: noqa: RUF001


def recursive_difference(
    dct1: dict | Any, dct2: dict | Any, cmprules: dict[type, callable] | None = None
) -> dict:
    """Return the recursive difference between two dicts.

    This is an *asymmetric* difference, i.e. we only end with values from dct1, and
    only if it is not exactly the same as in dct2. Entries in dct2 that are not in
    dct1 are ignored.

    Only dictionaries (and their subclasses) are recursed -- all other containers are
    treated as primitive types to be compared. If you need to recurse into custom
    structures, first "unstructure" your objects into dictionaries (e.g. with
    ``attrs.asdict(obj)``) then use this function.
    """
    cmprules = cmprules or {}

    out = {}
    for key, val1 in dct1.items():
        if isinstance(val1, dict):
            if not isinstance(dct2.get(key, None), dict):
                diff = val1
            else:
                diff = recursive_difference(val1, dct2[key], cmprules)
            if diff:
                out[key] = diff
        else:
            _eq = cmprules.get(type(val1), eq)
            if not _eq(val1, dct2.get(key, None)):
                out[key] = val1

    return out


def show_references(inputs: InputParameters, print_to_stdout=True):
    """Print out all the relevant references for a simulation based on input parameters."""
    ref_string = (
        "The main reference for 21cmFAST v3+:\n"
        "====================================\n"
        "Murray et al., (2020). 21cmFAST v3: A Python-integrated C code for generating 3D\n"
        "realizations of the cosmic 21cm signal. Journal of Open Source Software, 5(54),\n"
        "2582, https://doi.org/10.21105/joss.02582\n\n"
        "Based on the original 21cmFAST model:\n"
        "=====================================\n"
        'Andrei Mesinger, Steven Furlanetto and Renyue Cen, "21CMFAST: a fast, seminumerical\n'
        'simulation of the high-redshift 21-cm signal", Monthly Notices of the Royal\n'
        "Astronomical Society, Volume 411, Issue 2, pp. 955-972 (2011),\n"
        "https://ui.adsabs.harvard.edu/link_gateway/2011MNRAS.411..955M/doi:10.1111/j.1365-2966.2010.17731.x\n\n"
    )

    if inputs.matter_options.SOURCE_MODEL in [
        "E-INTEGRAL",
        "L-INTEGRAL",
        "DEXM-ESF",
        "CHMF-SAMPLER",
    ]:
        ref_string += (
            "The mass-dependent ionising efficiency model:\n"
            "=============================================\n"
            "Park, J., Mesinger, A., Greig, B., and Gillet, N.,\n"
            "“Inferring the astrophysics of reionization and cosmic dawn from galaxy luminosity\n"
            "functions and the 21-cm signal”, Monthly Notices of the Royal Astronomical Society,\n"
            "vol. 484, no. 1, pp. 933–949, 2019. https://doi.org/10.1093/mnras/stz032.\n\n"
        )

    if inputs.astro_options.USE_MINI_HALOS:
        ref_string += (
            "The minihalo model was first introduced in:\n"
            "===========================================\n"
            "Qin, Y., Mesinger, A., Park, J., Greig, B., and Muñoz, J. B.,\n"
            "“A tale of two sites - I. Inferring the properties of minihalo-hosted galaxies from\n"
            "current observations”, Monthly Notices of the Royal Astronomical Society, vol. 495,\n"
            "no. 1, pp. 123–140, 2020. https://doi.org/10.1093/mnras/staa1131.\n\n"
            "With improvements and DM-baryon relative velocity model in:\n"
            "===========================================================\n"
            "Muñoz, J.B., Qin, Y., Mesinger, A., Murray, S., Greig, B., and Mason, C.,\n"
            '"The Impact of the First Galaxies on Cosmic Dawn and Reionization",\n'
            "Monthly Notices of the Royal Astronomical Society, vol. 511, no. 3,\n"
            "pp 3657-3681, 2022 https://doi.org/10.1093/mnras/stac185\n"
        )

    if inputs.matter_options.lagrangian_source_grid:
        ref_string += (
            "Version 4 and the discrete halo sampler is described in:\n"
            "========================================================\n"
            "Davies, J. E., Mesinger, A., Murray, S. G.,\n"
            '"Efficient simulation of discrete galaxy populations and associated radiation\n'
            'fields during the first billion years",\n'
            "eprint arXiv:2504.17254, 2025. https://doi.org/10.48550/arXiv.2504.17254\n\n"
        )

    if print_to_stdout:
        print(ref_string)  # noqa: T201

    return ref_string



"""Contains the tests for the Perturbation algorithm (Linear, Zel'dovich, 2LPT).

Including perturbation of galaxy properties
"""

import numpy as np
import pytest

from py21cmfast import (
    InitialConditions,
    compute_halo_grid,
    perturb_field,
)
from py21cmfast.wrapper import cfuncs as cf


class TestPerturb:
    """Tests regarding the perturbation algorithms."""

    @pytest.fixture(scope="class")
    def test_pt_z(self):
        """Set redshift at which to test the 2LPT."""
        return 8.0

    @pytest.fixture(scope="class")
    def inputs_low(self, default_input_struct_ts):
        """Parameters for 2LPT tests."""
        # using 3-1 ratio for testing
        return default_input_struct_ts.evolve_input_structs(
            DIM=12,
            HII_DIM=4,
            BOX_LEN=8,
            SOURCE_MODEL="L-INTEGRAL",
            PERTURB_ON_HIGH_RES=False,
            R_BUBBLE_MAX=1.0,
        )

    @pytest.fixture(scope="class")
    def inputs_zel(self, inputs_low):
        """Parameters for Zel'dovich test."""
        return inputs_low.evolve_input_structs(
            PERTURB_ALGORITHM="ZELDOVICH",
        )

    @pytest.fixture(scope="class")
    def inputs_linear(self, inputs_low):
        """Parameters for Linear test."""
        return inputs_low.evolve_input_structs(
            PERTURB_ALGORITHM="LINEAR",
        )

    def get_fake_ics(self, inputs, test_pt_z):
        """Make an IC instance for the testing.

        These are inconsistent and strange values for real ICS but
        very trackable.
        """
        ics = InitialConditions.new(inputs=inputs)
        d_z = cf.get_growth_factor(inputs=inputs, redshift=test_pt_z)
        d_z_i = cf.get_growth_factor(
            inputs=inputs, redshift=inputs.simulation_options.INITIAL_REDSHIFT
        )

        res_fac = int(inputs.simulation_options.HIRES_TO_LOWRES_FACTOR)
        lo_dim = inputs.simulation_options.HII_DIM
        hi_dim = inputs.simulation_options.DIM
        fac_1lpt = inputs.simulation_options.cell_size / (d_z - d_z_i)
        fac_2lpt = inputs.simulation_options.cell_size / (
            (-3.0 / 7.0) * (d_z**2 - d_z_i**2)
        )
        for name, array in ics.arrays.items():
            setattr(ics, name, array.initialize().computed())

        # setup the velocities
        # NOTE: IC velocities are in Mpc
        if not inputs.matter_options.PERTURB_ON_HIGH_RES:
            fake_v = np.ones_like(ics.get("lowres_vx"))
            ics.set("lowres_vx", 0 * fake_v)
            ics.set("lowres_vy", fac_1lpt * fake_v)
            ics.set("lowres_vz", 0 * fake_v)
            if inputs.matter_options.PERTURB_ALGORITHM == "2LPT":
                ics.set("lowres_vx_2LPT", 0 * fake_v)
                ics.set("lowres_vy_2LPT", 0 * fake_v)
                ics.set("lowres_vz_2LPT", fac_2lpt * fake_v)
        else:
            fake_v = np.ones_like(ics.get("hires_vx"))
            ics.set("hires_vx", 0 * fake_v)
            ics.set("hires_vy", -fac_1lpt * fake_v)
            ics.set("hires_vz", 0 * fake_v)
            if inputs.matter_options.PERTURB_ALGORITHM == "2LPT":
                ics.set("hires_vx_2LPT", 0 * fake_v)
                ics.set("hires_vy_2LPT", 0 * fake_v)
                ics.set("hires_vz_2LPT", -fac_2lpt * fake_v)

        # set some densities that can be easily tracked
        d_lo = np.zeros_like(ics.get("lowres_density"))
        d_lo[0, 0, 0] = 1
        d_lo[lo_dim // 2, lo_dim // 2, lo_dim // 2] = -1
        ics.set("lowres_density", d_lo)
        # make similar hires densities
        d_hi = np.zeros_like(ics.get("hires_density"))
        d_hi[0, 0, 0] = res_fac**3
        d_hi[hi_dim // 2, hi_dim // 2, hi_dim // 2] = -(res_fac**3)
        ics.set("hires_density", d_hi)

        return ics

    @pytest.mark.parametrize("inputs", ["inputs_low", "inputs_zel"])
    def test_lowres_perturb(self, inputs, test_pt_z, request):
        """Tests low-resolution perturbation."""
        inputs = request.getfixturevalue(inputs)
        ics = self.get_fake_ics(inputs, test_pt_z)
        z_d = (
            test_pt_z
            if inputs.matter_options.PERTURB_ALGORITHM == "LINEAR"
            else inputs.simulation_options.INITIAL_REDSHIFT
        )
        roll_var = {
            "LINEAR": (0, 0, 0),
            "ZELDOVICH": (0, 1, 0),
            "2LPT": (0, 1, -1),
        }[inputs.matter_options.PERTURB_ALGORITHM]
        d_z = cf.get_growth_factor(inputs=inputs, redshift=z_d)

        # since we set the IC y,z velocities to move exactly one HII_DIM cell,
        # The expected density is the IC density rolled by one cell on two axes
        expected_dens = np.roll(ics.get("lowres_density"), roll_var, (0, 1, 2))
        expected_dens *= d_z
        pt = perturb_field(
            initial_conditions=ics,
            redshift=test_pt_z,
            regenerate=True,
            write=False,
        )
        np.testing.assert_allclose(pt.get("density"), expected_dens, atol=1e-3)

    @pytest.mark.skip(
        reason="aliasing in downsampling makes hires 2lpt unit tests difficult"
    )
    def test_hires_perturb(self, inputs_hi, test_pt_z):
        """Tests the high resolution perturbation."""
        ics = self.get_fake_ics(inputs_hi, test_pt_z)
        expected_dens = np.roll(ics.get("lowres_density"), (0, -1, 1), (0, 1, 2))
        d_z_i = cf.get_growth_factor(inputs=inputs_hi, redshift=test_pt_z)
        expected_dens *= d_z_i
        pt = perturb_field(
            initial_conditions=ics,
            redshift=test_pt_z,
            regenerate=True,
            write=False,
        )
        np.testing.assert_allclose(pt.get("density"), expected_dens, atol=1e-3)

    # TODO: include minihalo properties
    # TODO: include linear (for some reason)
    @pytest.mark.parametrize("inputs", ["inputs_low", "inputs_zel"])
    def test_hb_perturb(self, inputs, test_pt_z, request):
        """Tests the halo property perturbation."""
        inputs = request.getfixturevalue(inputs)
        ics = self.get_fake_ics(inputs, test_pt_z)
        hbox = compute_halo_grid(
            redshift=test_pt_z,
            initial_conditions=ics,
            inputs=inputs,
        )
        cell_radius = 0.620350491 * (
            inputs.simulation_options.BOX_LEN / inputs.simulation_options.HII_DIM
        )
        d_z = cf.get_growth_factor(
            inputs=inputs,
            redshift=test_pt_z,
        )
        roll_var = {
            "LINEAR": (0, 0, 0),
            "ZELDOVICH": (0, 1, 0),
            "2LPT": (0, 1, -1),
        }[inputs.matter_options.PERTURB_ALGORITHM]
        dens = np.roll(ics.get("lowres_density"), roll_var, (0, 1, 2)) * d_z
        mt_grid = np.full_like(dens, inputs.astro_params.M_TURN)

        prefac_sfr = (
            inputs.cosmo_params.cosmo.critical_density(0).to("Msun Mpc-3").value
            * inputs.astro_params.cdict["F_STAR10"]
            * inputs.cosmo_params.OMb
            * inputs.cosmo_params.cosmo.H(test_pt_z).to("s-1").value
            / inputs.astro_params.t_STAR
        )
        prefac_nion = (
            inputs.cosmo_params.cosmo.critical_density(0).to("Msun Mpc-3").value
            * inputs.astro_params.cdict["F_STAR10"]
            * inputs.cosmo_params.OMb
            * inputs.astro_params.cdict["F_ESC10"]
            * inputs.astro_params.cdict["POP2_ION"]
        )
        prefac_xray = (
            inputs.cosmo_params.cosmo.critical_density(0).to("Msun Mpc-3").value
            * inputs.cosmo_params.OMm
        )
        integral_sfrd, _ = cf.evaluate_SFRD_cond(
            inputs=inputs,
            redshift=test_pt_z,
            radius=cell_radius,
            densities=dens,
            log10mturns=mt_grid,
        )
        integral_sfrd *= prefac_sfr

        integral_nion, _ = cf.evaluate_Nion_cond(
            inputs=inputs,
            redshift=test_pt_z,
            radius=cell_radius,
            densities=dens,
            l10mturns_acg=mt_grid,
            l10mturns_mcg=mt_grid,
        )
        integral_nion *= prefac_nion

        integral_xray = cf.evaluate_Xray_cond(
            inputs=inputs,
            redshift=test_pt_z,
            radius=cell_radius,
            densities=dens,
            log10mturns=mt_grid,
        )
        integral_xray *= prefac_xray

        rtol = 1e-2
        np.testing.assert_allclose(hbox.get("halo_sfr"), integral_sfrd, rtol=rtol)
        np.testing.assert_allclose(hbox.get("n_ion"), integral_nion, rtol=rtol)
        np.testing.assert_allclose(hbox.get("halo_xray"), integral_xray, rtol=rtol)

    def test_hb_count_nonzero(self, inputs_low, test_pt_z):
        """Tests that the HaloBox count field is non-zero with EXTRA_HALOBOX_FIELDS=True."""
        from py21cmfast import config

        ics = self.get_fake_ics(inputs_low, test_pt_z)
        with config.use(EXTRA_HALOBOX_FIELDS=True):
            hbox = compute_halo_grid(
                redshift=test_pt_z,
                initial_conditions=ics,
                inputs=inputs_low,
            )
        count = hbox.get("count")
        assert count is not None, (
            "count field should be populated when EXTRA_HALOBOX_FIELDS=True"
        )
        assert np.any(count > 0), "count field should be non-zero"



"""
Module for the photon conservation models.

The excursion set reionisation model applied in ionize_box does not conserve photons.
As a result there is an offset between the expected global ionized fraction and the value
calculated from the ionized bubble maps. These models apply approximate corrections in order
to bring the bubble maps more in line with the expected global values.

The application of the model is controlled by the flag option PHOTON_CONS_TYPE, which
can take 4 values:

0. No correction is applied
1. We use the ionizing emissivity grids from a different redshift when calculating
   ionized bubble maps, this mapping from one redshift to another is obtained by
   performing a calibration simulation and measuring its redshift difference with the
   expected global evolution.
2. The power-law slope of the ionizing escape fraction is adjusted, using a fit
   ALPHA_ESC -> X + Y*Q(z), where Q is the expected global ionized fraction. This
   relation is fit by performing a calibration simulation as in (1), and comparing it to
   a range of expected global histories with different power-law slopes.
3. The normalisation of the ionizing escape fraction is adjusted, using a fit
   F_ESC10 -> X + Y*Q(z), where Q is the expected global ionized fraction. This relation
   is fit by performing a calibration simulation as in (1), and taking its ratio with
   the expected global evolution.

Notes
-----
The function map for the photon conservation model looks like::

    wrapper.run_lightcone/coeval()
        setup_photon_cons_correction()
            calibrate_photon_cons_correction()
                _init_photon_conservation_correction() --> computes and stores global evolution
                -->perfoms calibration simulation
                _calibrate_photon_conservation_correction() --> stores calibration evolution
            IF PHOTON_CONS_TYPE=='z-photoncons':
                lib.ComputeIonizedBox()
                    lib.adjust_redshifts_for_photoncons()
                        lib.determine_deltaz_for_photoncons() (first time only) --> calculates the deltaz array with some smoothing
                        --> does more smoothing and returns the adjusted redshift
            ELIF PHOTON_CONS_TYPE=='alpha-photoncons':
                photoncons_alpha() --> computes and stores ALPHA_ESC shift vs global neutral fraction
                lib.ComputeIonizedBox()
                    get_fesc_fit() --> applies the fit to the current redshift
            ELIF PHOTON_CONS_TYPE=='f-photoncons':
                photoncons_fesc() --> computes and stores F_ESC10 shift vs global neutral fraction
                lib.ComputeIonizedBox()
                    get_fesc_fit() --> applies the fit to the current redshift

"""

import logging

import attrs
import numpy as np
from scipy.optimize import curve_fit

from ..c_21cmfast import ffi, lib
from ._utils import _process_exitcode
from .cfuncs import broadcast_params
from .inputs import InputParameters
from .outputs import InitialConditions

logger = logging.getLogger(__name__)


# NOTE: if/when we move the photoncons model data to python we should use this
#   structure to hold z, xHI and parameter delta arrays
@attrs.define(kw_only=True)
class _PhotonConservationState:
    """Singleton class which contains the state of the photon-conservation model."""

    calibration_inputs: InputParameters | None = None

    @property
    def c_memory_allocated(self) -> bool:
        """Whether the memory for the parameter shifts has been allocated in the backend."""
        return lib.photon_cons_allocated

    @c_memory_allocated.setter
    def c_memory_allocated(self, val):
        lib.photon_cons_allocated = ffi.cast("bool", val)


_photoncons_state = _PhotonConservationState()


@broadcast_params
def _init_photon_conservation_correction(*, inputs, **kwargs):
    # This function calculates the global expected evolution of reionisation and saves
    #   it to C global arrays z_Q and Q_value (as well as other non-global confusingly named arrays),
    #   and constructs a GSL interpolator z_at_Q_spline
    return lib.InitialisePhotonCons()


def _calibrate_photon_conservation_correction(
    *, redshifts_estimate, nf_estimate, NSpline
):
    # This function passes the calibration simulation results to C,
    #       Storing a clipped version in global arrays nf_vals and z_vals,
    #       and constructing the GSL interpolator z_NFHistory_spline
    redshifts_estimate = np.array(redshifts_estimate, dtype="float64")
    nf_estimate = np.array(nf_estimate, dtype="float64")

    z = ffi.cast("double *", ffi.from_buffer(redshifts_estimate))
    xHI = ffi.cast("double *", ffi.from_buffer(nf_estimate))

    logger.debug(f"PhotonCons nf estimates: {nf_estimate}")
    return lib.PhotonCons_Calibration(z, xHI, NSpline)


def _calc_zstart_photon_cons():
    # gets the starting redshift of the z-based photon conservation model
    #   Set by neutral fraction astro_params.PHOTONCONS_ZSTART
    from ._utils import _call_c_simple

    return _call_c_simple(lib.ComputeZstart_PhotonCons)


def _get_photon_nonconservation_data() -> dict:
    """
    Access C global data representing the photon-nonconservation corrections.

    .. note::  If photon conservation is switched off via PHOTON_CONS_TYPE='no-photoncons' or the
               initialisation for photon conservation has not been performed yet, this
               will return None.

    Returns
    -------
    dict :
      z_analytic: array of redshifts defining the analytic ionized fraction
      Q_analytic: array of analytic  ionized fractions corresponding to `z_analytic`
      z_calibration: array of redshifts defining the ionized fraction from 21cmFAST without
      recombinations
      nf_calibration: array of calibration ionized fractions corresponding to `z_calibration`
      delta_z_photon_cons: the change in redshift required to calibrate 21cmFAST, as a function
      of z_calibration
      nf_photoncons: the neutral fraction as a function of redshift
    """
    # Check if photon conservation has been initialised at all
    if not _photoncons_state.c_memory_allocated:
        return {}

    arbitrary_large_size = 2000

    data = np.zeros((6, arbitrary_large_size))

    IntVal1 = np.array(np.zeros(1), dtype="int32")
    IntVal2 = np.array(np.zeros(1), dtype="int32")
    IntVal3 = np.array(np.zeros(1), dtype="int32")

    c_z_at_Q = ffi.cast("double *", ffi.from_buffer(data[0]))
    c_Qval = ffi.cast("double *", ffi.from_buffer(data[1]))
    c_z_cal = ffi.cast("double *", ffi.from_buffer(data[2]))
    c_nf_cal = ffi.cast("double *", ffi.from_buffer(data[3]))
    c_PC_nf = ffi.cast("double *", ffi.from_buffer(data[4]))
    c_PC_deltaz = ffi.cast("double *", ffi.from_buffer(data[5]))

    c_int_NQ = ffi.cast("int *", ffi.from_buffer(IntVal1))
    c_int_NC = ffi.cast("int *", ffi.from_buffer(IntVal2))
    c_int_NP = ffi.cast("int *", ffi.from_buffer(IntVal3))

    # Run the C code
    errcode = lib.ObtainPhotonConsData(
        c_z_at_Q,
        c_Qval,
        c_int_NQ,
        c_z_cal,
        c_nf_cal,
        c_int_NC,
        c_PC_nf,
        c_PC_deltaz,
        c_int_NP,
    )

    _process_exitcode(errcode, lib.ObtainPhotonConsData, ())

    ArrayIndices = [
        IntVal1[0],
        IntVal1[0],
        IntVal2[0],
        IntVal2[0],
        IntVal3[0],
        IntVal3[0],
    ]

    data_list = [
        "z_analytic",
        "Q_analytic",
        "z_calibration",
        "nf_calibration",
        "nf_photoncons",
        "delta_z_photon_cons",
    ]

    return {
        name: d[:index]
        for name, d, index in zip(data_list, data, ArrayIndices, strict=True)
    }


def setup_photon_cons(
    initial_conditions: InitialConditions,
    inputs: InputParameters | None = None,
    **kwargs,
):
    r"""
    Set up the photon non-conservation correction.

    First performs a simplified calibration simulation and saves its neutral fraction history,
    to be matched to the analytic expression from solving the filling factor
    ODE.

    This matching can happen via a redshift adjustment, or an adjustment to the escape fraction
    power-law parameters


    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions`,
        The InitialConditions instance to use for the photonconservation calculation
    inputs : :class:`~InputParameters`, optional
        An InputParameters instance. If not given will taken from initial_conditions.

    Other Parameters
    ----------------
    Any other parameters able to be passed to :func:`compute_initial_conditions`.
    """
    if inputs is None:
        inputs = initial_conditions.inputs

    if inputs.astro_options.PHOTON_CONS_TYPE == "no-photoncons":
        return {}

    from ..drivers._param_config import check_consistency_of_outputs_with_inputs

    check_consistency_of_outputs_with_inputs(inputs, [initial_conditions])

    # calculate global and calibration simulation xH histories and save them in C
    calibrate_photon_cons(
        inputs=inputs,
        initial_conditions=initial_conditions,
        **kwargs,
    )
    _photoncons_state.calibration_inputs = inputs

    # The PHOTON_CONS_TYPE == 1 case is handled in C (for now....), but we get the data anyway
    if inputs.astro_options.PHOTON_CONS_TYPE == "z-photoncons":
        photoncons_data = _get_photon_nonconservation_data()

    if inputs.astro_options.PHOTON_CONS_TYPE == "alpha-photoncons":
        photoncons_data = photoncons_alpha(inputs, **kwargs)

    if inputs.astro_options.PHOTON_CONS_TYPE == "f-photoncons":
        photoncons_data = photoncons_fesc(inputs)

    return photoncons_data


def calibrate_photon_cons(
    inputs: InputParameters,
    initial_conditions: InitialConditions,
    **kwargs,
):
    r"""
    Perform a photon conservation calibration simulation.

    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions`,
        The InitialConditions instance to use for the photonconservation calculation
    inputs : :class:`~InputParameters`,
        An InputParameters instance.

    Other Parameters
    ----------------
    See docs of :func:`compute_initial_conditions` for more information.
    """
    # avoiding circular imports by importing here
    from ..drivers.single_field import compute_ionization_field, perturb_field

    # Create a new astro_params and astro_options just for the photon_cons correction
    # NOTE: Since the calibration cannot do INHOMO_RECO, we set the R_BUBBLE_MAX
    #   to the default w/o recombinations ONLY when the original box has INHOMO_RECO enabled.
    # TODO: figure out if it's possible to find a "closest" Rmax, since the correction fails when
    # the histories are too different.

    # Using the halo sampling twice would be very slow, so switch to L-INTEGRAL for the calibration
    source_model_calibration = {
        "E-INTEGRAL": "E-INTEGRAL",
        "L-INTEGRAL": "L-INTEGRAL",
        "DEXM-ESF": "L-INTEGRAL",
        "CHMF-SAMPLER": "L-INTEGRAL",
    }
    inputs_calibration = inputs.evolve_input_structs(
        USE_TS_FLUCT=False,
        INHOMO_RECO=False,
        USE_MINI_HALOS=False,
        SOURCE_MODEL=source_model_calibration[inputs.matter_options.SOURCE_MODEL],
        PHOTON_CONS_TYPE="no-photoncons",
        R_BUBBLE_MAX=(
            15 if inputs.astro_options.INHOMO_RECO else inputs.astro_params.R_BUBBLE_MAX
        ),
    )
    ib = None
    prev_perturb = None

    # Arrays for redshift and neutral fraction for the calibration curve
    neutral_fraction_photon_cons = []

    # Initialise the analytic expression for the reionisation history
    logger.info("About to start photon conservation correction")
    _init_photon_conservation_correction(inputs=inputs, **kwargs)
    # Determine the starting redshift to start scrolling through to create the
    # calibration reionisation history
    logger.info("Calculating photon conservation zstart")
    z = _calc_zstart_photon_cons()

    fast_node_redshifts = [z]

    # NOTE: Not checking any redshift consistency for the calibration run
    #   Since the z-step is Q-dependent, we can't predict the redshifts
    inputs_calibration = inputs_calibration.clone(node_redshifts=None)

    while z > inputs.astro_params.PHOTONCONS_CALIBRATION_END:
        # Determine the ionisation box with recombinations, spin temperature etc.
        # turned off.
        this_perturb = perturb_field(
            redshift=z,
            inputs=inputs_calibration,
            initial_conditions=initial_conditions,
            **kwargs,
        )

        ib2 = compute_ionization_field(
            inputs=inputs_calibration,
            previous_ionized_box=ib,
            initial_conditions=initial_conditions,
            perturbed_field=this_perturb,
            previous_perturbed_field=prev_perturb,
            **kwargs,
        )

        mean_nf = np.mean(ib2.get("neutral_fraction"))

        # Save mean/global quantities
        neutral_fraction_photon_cons.append(mean_nf)

        # Can speed up sampling in regions where the evolution is slower
        if 0.3 < mean_nf <= 0.9:
            z -= 0.15
        elif 0.01 < mean_nf <= 0.3:
            z -= 0.05
        else:
            z -= 0.5

        ib = ib2
        if inputs.astro_options.USE_MINI_HALOS:
            prev_perturb = this_perturb

        fast_node_redshifts.append(z)

    fast_node_redshifts = np.array(fast_node_redshifts[::-1])
    neutral_fraction_photon_cons = np.array(neutral_fraction_photon_cons[::-1])

    # Construct the spline for the calibration curve
    logger.info("Calibrating photon conservation correction")
    _calibrate_photon_conservation_correction(
        redshifts_estimate=fast_node_redshifts,
        nf_estimate=neutral_fraction_photon_cons,
        NSpline=len(fast_node_redshifts),
    )


# (Jdavies): I needed a function to access the delta z from the wrapper
# get_photoncons_data does not have the edge cases that adjust_redshifts_for_photoncons does
@broadcast_params
def get_photoncons_dz(inputs, redshift, **kwargs):
    """Access the delta z arrays from the photon conservation model in C."""
    deltaz = np.zeros(1).astype("f4")
    redshift_pc_in = np.array([redshift]).astype("f4")
    stored_redshift_pc_in = np.array([redshift]).astype("f4")
    lib.adjust_redshifts_for_photoncons(
        ffi.cast("float *", redshift_pc_in.ctypes.data),
        ffi.cast("float *", stored_redshift_pc_in.ctypes.data),
        ffi.cast("float *", deltaz.ctypes.data),
    )

    return redshift_pc_in[0], stored_redshift_pc_in[0], deltaz[0]


# NOTE: alpha_func here MUST MATCH the C version TODO: remove one of them
def alpha_func(Q, a_const, a_slope):
    """Linear Function to fit in the simpler photon conservation model."""
    return a_const + a_slope * Q


# (jdavies): this will be a very hacky way to make a (d_alphastar vs z) array
# for a photoncons done by ALPHA_STAR instead of redshift
# This will work by taking the calibration simulation, plotting a RANGE of analytic
# Q vs z curves for different ALPHA_STAR, and then finding the aloha star which has the inverse ratio
# with the reference analytic as the calibration
# TODO: don't rely on the photoncons functions since they do a bunch of other stuff in C
def photoncons_alpha(inputs, **kwargs):
    """Run the Simpler photons conservation model using ALPHA_ESC.

    This adjusts the slope of the escape fraction instead of redshifts to match a global
    evolution.
    """
    # HACK: I need to allocate the deltaz arrays so I can return the other ones properly, this isn't a great solution
    # TODO: Move the deltaz interp tables to python
    if not _photoncons_state.c_memory_allocated:
        lib.determine_deltaz_for_photoncons()
        _photoncons_state.c_memory_allocated = True

    # Q(analytic) limits to fit the curve
    max_q_fit = 0.99
    min_q_fit = 0.2

    ap_c = inputs.astro_params.cdict

    ref_pc_data = _get_photon_nonconservation_data()
    z = ref_pc_data["z_calibration"]
    alpha_arr = (
        np.linspace(-2.0, 1.0, num=31) + ap_c["ALPHA_ESC"]
    )  # roughly -0.1 steps for an extended range of alpha
    test_pc_data = np.zeros((alpha_arr.size, ref_pc_data["z_calibration"].size))

    # fit to the same z-array
    ref_interp = np.interp(
        ref_pc_data["z_calibration"],
        ref_pc_data["z_analytic"],
        ref_pc_data["Q_analytic"],
    )
    for i, a in enumerate(alpha_arr):
        # alter astro params with new alpha
        inputs_photoncons = inputs.evolve_input_structs(ALPHA_ESC=a)

        # find the analytic curve wth that alpha
        # TODO: Theres a small memory leak here since global arrays are allocated (for some reason)
        # TODO: use ffi to free them?
        #       This will be fixed by moving the photoncons to python
        _init_photon_conservation_correction(inputs=inputs_photoncons, **kwargs)

        # save it
        pcd_buf = _get_photon_nonconservation_data()

        # interpolate to the calibration redshifts
        test_pc_data[i, ...] = np.interp(
            ref_pc_data["z_calibration"], pcd_buf["z_analytic"], pcd_buf["Q_analytic"]
        )

    # filling factors sometimes go above 1, this causes problems in late-time ratios
    # I want this in the test alphas to get the right photon ratio, but not in the reference analytic
    ref_interp[ref_interp > 1] = 1.0

    # ratio of each alpha with calibration
    ratio_test = (test_pc_data) / ref_interp[None, ...]
    # ratio of given alpha with calibration
    ratio_ref = (1 - ref_pc_data["nf_calibration"]) / ref_interp

    ratio_diff = ratio_test - 1 / ratio_ref[None, :]  # find N(alpha)/ref == ref/cal
    diff_test = (
        (test_pc_data)
        + (1 - ref_pc_data["nf_calibration"])[None, ...]
        - 2 * ref_interp[None, ...]
    )  # find N(alpha) - ref == ref - cal
    reverse_test = (
        test_pc_data - (1 - ref_pc_data["nf_calibration"])[None, ...]
    )  # find NF(alpha) == cal, then apply - alpha

    alpha_estimate_ratio = np.zeros_like(z)
    alpha_estimate_diff = np.zeros_like(z)
    alpha_estimate_reverse = np.zeros_like(z)

    # find the roots of each function
    roots_ratio_idx = np.where(np.diff(np.sign(ratio_diff), axis=0))
    roots_diff_idx = np.where(np.diff(np.sign(diff_test), axis=0))
    roots_reverse_idx = np.where(np.diff(np.sign(reverse_test), axis=0))

    # find alpha estimates
    for arr_in, roots_arr, arr_out in zip(
        [ratio_diff, diff_test, reverse_test],
        [roots_ratio_idx, roots_diff_idx, roots_reverse_idx],
        [alpha_estimate_ratio, alpha_estimate_diff, alpha_estimate_reverse],
        strict=True,
    ):
        last_alpha = ap_c["ALPHA_ESC"]
        for i in range(z.size)[::-1]:
            # get the roots at this redshift
            roots_z = np.where(roots_arr[1] == i)
            # if there are no roots, assign nan
            if roots_z[0].size == 0:
                arr_out[i] = np.nan
                continue

            alpha_idx = roots_arr[0][roots_z]

            # interpolate
            y0 = arr_in[alpha_idx, i]
            y1 = arr_in[alpha_idx + 1, i]
            x0 = alpha_arr[alpha_idx]
            x1 = alpha_arr[alpha_idx + 1]
            guesses = -y0 * (x1 - x0) / (y1 - y0) + x0

            # choose the root which gives the smoothest alpha vs z curve
            arr_out[i] = guesses[np.argmin(np.fabs(guesses - last_alpha))]
            last_alpha = arr_out[i]

    # initialise the output structure before the fits
    results = {
        "z_calibration": ref_pc_data["z_calibration"],
        "z_analytic": ref_pc_data["z_analytic"],
        "Q_analytic": ref_pc_data["Q_analytic"],
        "nf_photoncons": 1 - ref_interp,
        "Q_alpha": test_pc_data,
        "nf_calibration": ref_pc_data["nf_calibration"],
        "alpha_ratio": alpha_estimate_ratio,
        "alpha_diff": alpha_estimate_diff,
        "alpha_reverse": alpha_estimate_reverse,
        "alpha_arr": alpha_arr,
        "fit_yint": ap_c["ALPHA_ESC"],
        "fit_slope": 0,  # start with no correction
        "found_alpha": False,
    }

    # adjust the reverse one (we found the alpha which is close to the calibration sim, undo it)
    alpha_estimate_reverse = 2 * ap_c["ALPHA_ESC"] - alpha_estimate_reverse

    # fit to the curve
    # make sure there's an estimate and Q isn't too high/low
    fit_alpha = alpha_estimate_ratio
    sel = np.isfinite(fit_alpha) & (ref_interp < max_q_fit) & (ref_interp > min_q_fit)

    # if there are no alpha roots found, it's likely this is a strange reionisation history
    # but we can't apply the alpha correction so throw an error

    # if we can't fit due to not enough ionisation, catch that here
    if ref_interp.max() < min_q_fit:
        results["fit_yint"] = last_alpha
        logger.warning(
            f"These parameters result in little ionisation, running with flat alpha correction {last_alpha}"
        )

    elif np.count_nonzero(sel) == 1:
        results["fit_yint"] = last_alpha
        logger.warning(
            f"ONLY ONE REDSHIFT HAD AN ALPHA FIT WITHIN THE RANGE, running with flat alpha correction {last_alpha}"
        )

    # here there are no fits, meaning we likely have a strange reionisaiton history where the photon conservation is worse than alpha can correct
    elif np.count_nonzero(sel) == 0:
        logger.warning(
            "No alpha within the range can fit the global history for any z, running without alpha correction"
        )
        logger.warning(
            "THIS REIONISAITON HISTORY IS LIKELY HIGHLY NON-CONSERVING OF PHOTONS"
        )

    else:
        popt, _pcov = curve_fit(alpha_func, ref_interp[sel], fit_alpha[sel])
        # pass to C
        logger.info(f"ALPHA_ESC Original = {ap_c['ALPHA_ESC']:.3f}")
        logger.info(f"Running with ALPHA_ESC = {popt[0]:.2f} + {popt[1]:.2f} * Q")

        results["fit_yint"] = popt[0]
        results["fit_slope"] = popt[1]
        results["found_alpha"] = True

    lib.set_alphacons_params(results["fit_yint"], results["fit_slope"])

    return results


def photoncons_fesc(inputs):
    """Run the Even Simpler photon conservation model using F_ESC10.

    Adjusts the normalisation of the escape fraction to match a global evolution.
    """
    # HACK: I need to allocate the deltaz arrays so I can return the other ones properly, this isn't a great solution
    if not _photoncons_state.c_memory_allocated:
        lib.determine_deltaz_for_photoncons()
        _photoncons_state.c_memory_allocated = True

    # Q(analytic) limits to fit the curve
    max_q_fit = 0.99
    min_q_fit = 0.2

    ref_pc_data = _get_photon_nonconservation_data()

    # fit to the same z-array
    ref_interp = np.interp(
        ref_pc_data["z_calibration"],
        ref_pc_data["z_analytic"],
        ref_pc_data["Q_analytic"],
    )

    ap_c = inputs.astro_params.cdict
    # filling factors sometimes go above 1, this causes problems in late-time ratios
    # I want this in the test alphas to get the right photon ratio, but not in the reference analytic
    ref_interp[ref_interp > 1] = 1.0

    # ratio of each alpha with calibration
    ratio_ref = ref_interp / (1 - ref_pc_data["nf_calibration"])

    fit_fesc = ratio_ref * ap_c["F_ESC10"]
    sel = np.isfinite(fit_fesc) & (ref_interp < max_q_fit) & (ref_interp > min_q_fit)

    popt, _pcov = curve_fit(alpha_func, ref_interp[sel], fit_fesc[sel])
    # pass to C
    logger.info(f"F_ESC10 Original = {ap_c['F_ESC10']:.3f}")
    logger.info(f"Running with F_ESC10 = {popt[0]:.2f} + {popt[1]:.2f} * Q")

    # initialise the output structure before the fits
    results = {
        "z_calibration": ref_pc_data["z_calibration"],
        "z_analytic": ref_pc_data["z_analytic"],
        "Q_analytic": ref_pc_data["Q_analytic"],
        "nf_calibration": ref_pc_data["nf_calibration"],
        "nf_photoncons": 1 - ref_interp,
        "Q_ratio": ratio_ref,
        "fesc_target": fit_fesc,
        "fit_yint": popt[0],
        "fit_slope": popt[1],  # start with no correction
    }

    lib.set_alphacons_params(results["fit_yint"], results["fit_slope"])

    return results



"""Test initializing tables in C."""

from py21cmfast.c_21cmfast import lib

from py21cmfast.wrapper.cfuncs import broadcast_input_struct


def test_init_heat(default_input_struct):
    # NOTE: this is not ideal, but the tables depend on POPX_NION which is no longer global
    broadcast_input_struct(inputs=default_input_struct)
    out = lib.init_heat()
    assert out == 0
